"""
    SecTheory.MonoporeSecSimulation.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from scipy.optimize import minimize, basinhopping
from matplotlib.gridspec import GridSpec

def demo(use_cf=True):
    import molass_legacy.KekLib.DebugPlot as plt
    if use_cf:
        from SecTheory.SecCF import sdm_monopore
        from SecTheory.SecPDF import FftInvPdf
        monopore_pdf = FftInvPdf(sdm_monopore)
    else:
        # 
        assert False

    x = np.arange(300)

    me = 0.1
    mp = 1.5
    rp = 76
    rg = 35
    rho = rg/rp
    N = 171
    T = 0.63
    t0 = 100
    sigma0 = 10
    N0 = (t0/sigma0)**2
    x0 = -20

    rg_array = np.array([60, 50, 40, 30, 20, 10, 5])

    y_list = []
    for rg_ in rg_array:
        rho_ = rg_/rp
        np_ = N*(1 - rho_)**me
        tp_ = T*(1 - rho_)**mp
        if use_cf:
            y = monopore_pdf(x, np_, tp_, N0, t0, x0)
        else:
            assert False
        y_list.append(y)

    with plt.Dp():
        fig, ax = plt.subplots()

        for y, rg_ in zip(y_list, rg_array):
            ax.plot(x, y, label="$R_g=%.3g$" % rg_)

        ax.legend()
        fig.tight_layout()
        plt.show()

def compare_stochastic_dispersive():
    import molass_legacy.KekLib.DebugPlot as plt
    from SecTheory.SecCF import shifted_phi, sdm_monopore
    from SecTheory.SecPDF import FftInvPdf
    gec_monopore_pdf = FftInvPdf(shifted_phi)
    sdm_monopore_pdf = FftInvPdf(sdm_monopore)

    x = np.arange(300)

    me = 0.1
    mp = 1.5
    rp = 76
    rg = 35
    rho = rg/rp
    N = 171
    T = 0.63
    t0 = 100
    sigma0 = 10
    N0 = (t0/sigma0)**2
    x0 = 50

    rg_array = np.array([60, 50, 40, 30, 20, 10, 5])

    y1_list = []
    y2_list = []
    for rg_ in rg_array:
        rho_ = rg_/rp
        np_ = N*(1 - rho_)**me
        tp_ = T*(1 - rho_)**mp
        y1 = gec_monopore_pdf(x, np_, tp_, t0 - x0)
        y1_list.append(y1)
        y2 = sdm_monopore_pdf(x, np_, tp_, N0, t0, -x0)
        y2_list.append(y2)

    with plt.Dp():
        fig, (ax1,ax2) = plt.subplots(nrows=2, figsize=(8,10))

        ax1.set_title("Simple Monopore Model", fontsize=16)
        ax2.set_title("Dispeasive Monopore Model", fontsize=16)

        for y1, y2, rg_ in zip(y1_list, y2_list, rg_array):
            ax1.plot(x, y1, label="$R_g=%.3g$" % rg_)
            ax2.plot(x, y2, label="$R_g=%.3g$" % rg_)

        ax1.legend()
        ax2.legend()
        fig.tight_layout()
        plt.show()

def compare_simple_curve():
    import molass_legacy.KekLib.DebugPlot as plt
    from SecTheory.SecCF import sdm_monopore
    from SecTheory.SecPDF import FftInvPdf
    monopore_pdf = FftInvPdf(sdm_monopore)

    x = np.arange(300)

    me = 0.1
    mp = 1.5
    rp = 200
    rg = 35
    rho = rg/rp
    N = 171
    T = 0.63
    t0 = 100
    sigma0 = 10
    N0 = (t0/sigma0)**2
    x0 = -20

    rg_array = np.array([180, 150, 120, 90, 80, 70, 60, 50, 40, 30, 20, 10])

    y_list = []
    p_list = []
    for rg_ in rg_array:
        rho_ = rg_/rp
        np_ = N*(1 - rho_)**me
        tp_ = T*(1 - rho_)**mp
        y = monopore_pdf(x, np_, tp_, N0, t0, x0)
        y_list.append(y)
        j = np.argmax(y)
        p_list.append((x[j], y[j]))

    r = np.linspace(0, rp, 100)
    rho = r/rp
    P = 300
    rho_ = rg_array/rp
    rho_[rho_ > 1] = 1
    p_array = np.array(p_list)
    x_array = p_array[:,0]

    m = me + mp

    def objective(p):
        t0_, P_ = p
        tr_ = t0_ + P_*(1 - rho_)**m
        return np.sum((tr_ - x_array)**2)

    # method="Nelder-Mead"
    bounds = [(None, None), (0, None)]
    ret = basinhopping(objective, (t0, P), minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds)
    )

    t0opt, Popt = ret.x
    tr = t0opt + Popt*(1 - rho)**m
    tr_ = t0opt + Popt*(1 - rho_)**m

    print("t0opt=", t0opt)
    print("Popt=", Popt)

    with plt.Dp():
        fig, ax = plt.subplots()

        ax.set_title(r"$ t_R = t_0 + P(1 - \rho)^m $ Fitting to Monopore Model", fontsize=16)

        axt = ax.twinx()
        axt.grid(False)

        for y, rg_ in zip(y_list, rg_array):
            ax.plot(x, y, label="$R_g=%.3g$" % rg_)

        ax.plot(*p_array.T, "o", color="yellow")

        axt.plot(tr, r, color="yellow")
        axt.plot(tr_, rg_array, "o", color="red")

        ymin, ymax = axt.get_ylim()
        axt.set_ylim(-ymax, ymax)

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    # demo()
    compare_stochastic_dispersive()
    # compare_simple_curve()



""" borrowd from:
    Plot a plane based on a normal vector and a point in Matlab or matplotlib
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

point  = np.array([1, 2, 3])
normal = np.array([1, 1, 2])

# a plane is a*x+b*y+c*z+d=0
# [a,b,c] is the normal. Thus, we have to calculate
# d and we're set
d = -point.dot(normal)

# create x,y
xx, yy = np.meshgrid(range(10), range(10))

print( 'xx=', xx )
print( 'yy=', yy )

# calculate corresponding z
z = (-normal[0] * xx - normal[1] * yy - d) * 1. /normal[2]

# plot the surface
plt3d = plt.figure().add_subplot(111, projection='3d')
plt3d.plot_surface(xx, yy, z, alpha=0.5 )
plt.show()



"""
    Estimators.SdmEstimator.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.RetensionTime import estimate_init_rgs
from molass_legacy.Peaks.PeProgressConstants import MAXNUM_STEPS, STOCH_INIT_STEPS
from .BaseEstimator import BaseEstimator

class SdmEstimator(BaseEstimator):
    def __init__(self, editor):
        BaseEstimator.__init__(self, editor)

    def estimate_params(self, lrf_src=None, edm_available=False, debug=False):
        return self.compute_sdm_init_params(self.nc, lrf_src=lrf_src, edm_available=edm_available, debug=debug)

    def compute_sdm_init_params(self, nc_b, lrf_src=None, edm_available=False, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.Models.Stochastic.DispersiveMonopore
            reload(molass_legacy.Models.Stochastic.DispersiveMonopore)
            import molass_legacy.Models.Stochastic.MonoporeUvScaler
            reload(molass_legacy.Models.Stochastic.MonoporeUvScaler)
            import molass_legacy.Estimators.SdmEstimatorImpl
            reload(molass_legacy.Estimators.SdmEstimatorImpl)
            import molass_legacy.Peaks.PeakFronting
            reload(molass_legacy.Peaks.PeakFronting)
        from molass_legacy.Models.Stochastic.DispersiveMonopore import guess_params_using_moments
        from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
        
        from .SdmEstimatorImpl import guess_exec_spec, edit_to_full_sdmparams
        from molass_legacy.Peaks.PeakFronting import has_fronting_peak
 
        editor = self.editor
        optimizer = editor.fullopt
    
        progress = MAXNUM_STEPS - STOCH_INIT_STEPS
        editor.update_status_bar("Estimating stochastic initial parameters.")
     
        nc = nc_b - 1   # num components without baseline
        (xr_curve, D), rg_curve, (uv_curve, U) = editor.dsets

        if lrf_src is None:
            lrf_src = editor.get_lrf_source(devel=True)
            if lrf_src is None:
                assert False, "No LRF source"

        info = lrf_src.compute_rgs(want_num_components=nc, debug=False)
        if info is None:
            return

        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = info
        peaks = lrf_src.xr_peaks[indeces]
        if edm_available:
            # removing the area M[0]
            egh_moments_list = [M[1:] for M in optimizer.compute_moments_list(debug=True)]
        else:
            egh_moments_list = compute_egh_moments(peaks)
        x = lrf_src.xr_x
        y = lrf_src.xr_y
        exec_spec = guess_exec_spec(peak_rgs, props, qualities)
        self.logger.info("init_params are estimated using exec_spec: %s", exec_spec)
        fronting = has_fronting_peak(xr_curve, debug=debug)
        if debug:
            print("indeces=", indeces, "qualities=", qualities)
            print("fronting=", fronting)                   
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("compute_sdm_init_params debug")
                if edm_available:
                    axt = ax.twinx()
                    axt.grid(False)
                    axis_info = (fig, (None, ax, None, axt))
                    optimizer.objective_func(optimizer.init_params, plot=True, axis_info=axis_info)
                else:
                    model = lrf_src.model
                    ax.plot(x, y, label="data")
                    for k, (params, rg, quality) in enumerate(zip(peaks, peak_rgs, qualities)):
                        cy = model(x, params)
                        ax.plot(x, cy, ":", label="Rg=%.3g (%.3g)" % (rg, quality))
                    ax.legend()
                fig.tight_layout()
                ret = plt.show()
            if not ret:
                return
        ret = guess_params_using_moments(x, y, egh_moments_list, peak_rgs, qualities, props,
                                            fronting=fronting,
                                            exec_spec=exec_spec, debug=debug)
        if ret is None:
            return
    
        progress += 1
        editor.pbar["value"] = progress
        editor.update()

        sdm_params, corrected_rgs, bounds = ret
        self.bounds = bounds

        init_params = edit_to_full_sdmparams(editor, sdm_params, corrected_rgs, uv_curve, debug=debug)
        if init_params is None:
            return

        progress += 1
        editor.pbar["value"] = progress
        editor.update_status_bar("Stochastic initial parameters are ready.")
        self.logger.info("init_params=%s", str(init_params))

        return init_params

    def get_colparam_bounds_bug(self):
        from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCOL_PARAMS

        
        est_col_bounds = list(self.bounds[0:NUM_SDMCOL_PARAMS])
        return est_col_bounds[0:4] + [(1600, 60000)] + est_col_bounds[4:]

    def get_colparam_bounds(self):
        # temporary fix for bug
        from molass_legacy.Models.Stochastic.ParamLimits import MNP_BOUNDS
        # MNP_BOUNDS = [N_BOUND, KT_BOUND, T0_BOUND, PORESIZE_BOUNDS]
        # N, K, x0, poresize, N0, tI = sdmcol_params
        mnp_bounds = MNP_BOUNDS.copy()
        return mnp_bounds + [(1600, 60000), (-1000, 0)]

def onthefly_test(editor):
    estimator = SdmEstimator(editor)
    print("estimating...")
    init_params = estimator.estimate_params(debug=True)
    print("done.")
    if init_params is not None:
        editor.fullopt.params_type.set_estimator(estimator)
        def components_plot_debug():
            from importlib import reload
            import molass_legacy.Estimators.SdmEstimatorDebug
            reload(molass_legacy.Estimators.SdmEstimatorDebug)
            from .SdmEstimatorDebug import components_plot_debug_impl
            components_plot_debug_impl(editor.fullopt, init_params)

        with plt.Dp(extra_button_specs=[("Components Plot Debug", components_plot_debug)]):
            fig, ax = plt.subplots()
            ax.set_title("components_plot_debug")
            ret = plt.show()
        if not ret:
            print("debug done.")
            return
        editor.draw_scores(init_params, create_new_optimizer=False)
        print("redraw done.")


"""

    ExcelEnvDepend.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
import os
import numpy as np
from time import sleep
from openpyxl import Workbook
from openpyxl.chart import LineChart, Reference
from molass_legacy._MOLASS.SerialSettings import get_setting

CHART_WIDTH     = 54

def add_a_chart( ws, name ):

    num_points = 160
    num_cols = 2

    title_row = []

    data = np.random.uniform(0, 1, (num_points, num_cols))

    for k in range(num_points):
        row = data[k,:]
        ws.append(list(row))

    c1 = LineChart()
    c1.title = name
    c1.style = 13
    c1.y_axis.title = 'Values'
    c1.x_axis.title = 'Eno'
    c1.width = CHART_WIDTH

    data = Reference(ws, min_col=1, min_row=1, max_col=num_cols, max_row=num_points)
    c1.add_data(data, titles_from_data=False)
    ws.add_chart(c1, "R10")

xlTypePDF = 0

def get_chart_width_adjust_ratio(excel_client):
    temp_folder = get_setting('temp_folder')
    book_path = os.path.join(temp_folder, "tempbook.xlsx")

    wb = Workbook()
    ws = wb.active
    add_a_chart(ws, "Test Chart")
    wb.save(book_path)
    wb.close()

    app = excel_client.excel

    ole_ws = excel_client.openWorksheet(book_path + "(Sheet)")
    top_chart = ole_ws.get_chart(0)
    top_area  = top_chart.ChartArea
    plot_area  = top_chart.PlotArea

    c_left = top_area.Left
    c_width = top_area.Width
    c_top = top_area.Top
    c_height = top_area.Height

    x_left = c_left
    x_right = c_left + c_width
    y_top = c_top
    y_btm = c_top + c_height

    line1 = ole_ws.draw_line( ( x_left,  y_top ), ( x_left,  y_btm ) )
    line2 = ole_ws.draw_line( ( x_right, y_top ), ( x_right, y_btm ) )

    for shape in line1, line2:
        shape.Select()
        line_ = excel_client.selection().ShapeRange.Line
        line_.ForeColor.RGB = 0x00FF00

    x_left = c_left + plot_area.Left
    x_right = x_left + plot_area.Width
    y_top = c_top
    y_btm = c_top + c_height

    line3 = ole_ws.draw_line( ( x_left,  y_top ), ( x_left,  y_btm ) )
    line4 = ole_ws.draw_line( ( x_right, y_top ), ( x_right, y_btm ) )

    for shape in line3, line4:
        shape.Select()
        line_ = excel_client.selection().ShapeRange.Line
        line_.ForeColor.RGB = 0xFF0000

    if False:

        app.Cells.Select()
        app.Selection.CopyPicture()
        active = app.Sheets.Add()
        active.Paste()

        # active.Pictures().Paste().Select()
        # active.Shapes.Range("Picture 1").Select()
        if True:
            img_path = os.path.join(temp_folder, "temp.jpg")
            active.Shapes(1).SaveAsPicture(img_path)
        else:
            pdf_path = os.path.join(temp_folder, "temp.pdf")
            active.ExportAsFixedFormat(xlTypePDF, pdf_path)

        # active.Pictures().SaveAsPicture(pic_path)
        # app.Selection.ExportAsFixedFormat(xlTypePDF, "temp.pdf")
        # print(active.Shapes.Count)
        # print(active.Shapes.Item(1))

        # active.Shapes.Item(1).SaveAsPicture("temp.jpg")
        # active.Shapes(1).SaveAsPicture("temp.jpg")
        # active.Shapes.Item(1).Chart.Export("temp.jpg")

    workbook = ole_ws.workbook
    workbook.Save()
"""
Sub Macro1()
'
' Macro1 Macro
'
    Cells.Select
    Application.Left = 174.25
    Application.Top = 25.75
    Selection.Copy
    Sheets.Add After:=ActiveSheet
    ActiveSheet.Pictures.Paste.Select
    ActiveSheet.Shapes.Range(Array("Picture 1")).Select
End Sub
"""



# coding: utf-8
"""
    ClientServer.py

    Copyright (c) 2016-2019, Masatsuyo Takahashi, KEK-PF
"""

import os
import socket
import logging
import itertools
import time
import warnings
from threading              import Thread
from molass_legacy.KekLib.BasicUtils             import mkdirs_with_retry
# from KekToolsGP             import AutorgKek
from molass_legacy.AutorgKekAdapter       import AutorgKekAdapter as AutorgKek
from ResultPrinter          import ResultPrinter
from Settings               import get_setting

DEBUG = False

HOST        = 'localhost'
PORT        = 49999
BUFFER_SIZE = 1024
STOP        = b'--stop_server'
EXECUTE     = b'EXECUTE'
ASK_READY   = b'ASK_READY'
ACK         = b'ACK'
QVERSION    = b'--version'
SUCCESS     = b'SUCCESS'
FAILURE     = b'FAILURE'
MESSAGE     = b'MESSAGE'
SERIAL_ANALYZER = b'SERIAL_ANALYZER'
PROCESS_ARGS    = b'PROCESS_ARGS'
START_PROCESS   = b'START_PROCESS'
GET_NUM_STEPS   = b'GET_NUM_STEPS'
STOP_PROCESS    = b'STOP_PROCESS'
REQUEST_INFO    = b'REQUEST_INFO'
CANCEL_PROCESS  = b'CANCEL_PROCESS'
UNEXPECTED      = b'UNEXPECTED'

exe_options = {
    '--ATSAS'   : ( 'ATSAS', True ),
    '--atsas'   : ( 'ATSAS', True ),
    '--robust'  : ( 'robust', True ),
    }

class AutorgKekServer:
    def __init__( self, command_obj, host=HOST, port=PORT ):
        self.cmmd   = command_obj
        self.host   = host
        self.port   = port
        self.thread_array = []

        updir   = os.path.abspath( __file__ )
        for i in range(3):
            updir   = os.path.dirname( updir )
        logdir  = updir + '/log'
        mkdirs_with_retry( logdir )
        logfile = logdir + '/autorg_kek_server-log.csv'
        bakfile = logfile.replace( '.csv', '.bak' )

        if os.path.exists( logfile ):
            if os.path.exists( bakfile ):
                os.remove( bakfile )
            os.rename( logfile, bakfile )

        logging.basicConfig(
            filename=logfile, filemode='w',
            format='%(asctime)s,%(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            level=logging.DEBUG,
            )

        self.logger = logging.getLogger( __name__ )
        self.logger.info( 'Server started' )

    def get_port( self ):
        return self.port

    def run( self ):
        warnings.filterwarnings("ignore")

        self.socket = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
        self.socket.bind( ( self.host, self.port ) )
        self.socket.listen( 1 )

        stop_server = False

        while not stop_server:
            conn, addr = self.socket.accept()
            if DEBUG: print( 'Connection address:', addr )

            args = []
            data = self.communicate( conn, args )
            if data == STOP:
                stop_server = True

            if data == SERIAL_ANALYZER:
                # conn will be closed in the thread
                self.start_serial_controller_thread( conn )

            else:
                while data != b'':
                    data = self.communicate( conn, args )
                    if data == STOP:
                        stop_server = True
                conn.close()

        self.logger.info( 'Server stopped' )

    def communicate( self, conn, args ):
        data = conn.recv( BUFFER_SIZE )
        if DEBUG: print( "server received data:", data )

        if data == b'':
            return data

        if data==ASK_READY:
            reply = ACK
        elif data == STOP:
            reply = ACK
        elif data == EXECUTE:
            reply = self.execute( args )
        elif data == QVERSION:
            from molass_legacy.AutorgKek.AppVersion import get_com_version_string
            reply = MESSAGE + bytes( ';' + get_com_version_string(), 'utf-8' )
        else:
            args.append( data.decode() )
            reply = ACK
        conn.send( reply )
        return data

    def execute( self, args ):
        # print( 'TO EXECUTE: args=', args )
        rq_args = args[0:2]
        self.log( rq_args )
        kw_args = {}
        for i, v in enumerate( args[2:] ):
            opt = exe_options.get( v )
            if opt is not None:
                kw_args[ opt[0] ] = opt[1]
        if DEBUG: print( 'TO EXECUTE: ', rq_args, kw_args )

        # ret = self.cmmd.execute( *rq_args, **kw_args )

        in_file, out_file = rq_args
        ATSAS       = False if len(args) <= 2 else args[2]
        robust      = True  if len(args) <= 3 else args[3]
        optimize    = True  if len(args) <= 4 else args[4]

        try:
            autorg = AutorgKek( in_file )
            result = autorg.run( robust=robust, optimize=True )
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.info( str(etb) )
            result = None

        if result is None: return FAILURE

        if robust or result.Rg is not None:
            fh = open( out_file, "w" )

            printer = ResultPrinter(
                fh = fh,
                csv_output = out_file.find( '.csv' ) > 0,
                ATSAS=ATSAS,
                )
            printer.print_result( 0, in_file, result )

            fh.close()

        ret = result

        if ret is None:
            Rg = None
        else:
            Rg = ret.Rg
        self.log( rq_args + [ 'NA' if Rg is None else '%g' % Rg ] )

        reply = FAILURE if ret is None else SUCCESS

        return reply

    def log( self, *args ):
        message = ','.join( list( itertools.chain( *args ) ) )
        self.logger.info( message )

    def start_serial_controller_thread( self, conn ):
        thread = SerialControllerThread( conn )
        thread.start()
        self.logger.info( 'Serial Analyzer started' )
        self.thread_array.append( thread )

class SerialControllerThread(Thread):
    def __init__( self, conn ):
        Thread.__init__( self )
        self.conn = conn

    def receive( self, decode=False ):
        data = self.conn.recv( BUFFER_SIZE )
        if DEBUG: print( "SerialControllerThread received data:", data )
        self.conn.send( ACK )
        if decode:
            return data.decode()
        else:
            return data

    def run( self ):
        from SerialController           import SerialController
        warnings.filterwarnings("ignore")
        if DEBUG: print( "SerialControllerThread.run" )
        args = []

        while True:

            try:
                if DEBUG: print( 'SerialControllerThread: recv at loop top' )

                data = self.receive()

                if DEBUG: print( 'SerialControllerThread: data=', data )
                if data == b'' or data == REQUEST_INFO:
                    break

            except Exception as exc:
                print( exc )
                break

            if data == PROCESS_ARGS:
                args = []
                for i in range(3):
                    args.append( self.receive( decode=True ) )
                continue

            elif data == STOP_PROCESS:
                break

            elif data == START_PROCESS:
                if len(args) != 3:
                    print( 'len(args) != 3:' )
                    break

                data_folder = args[0]
                conc_folder = args[1]
                outp_folder = args[2]
                result_book = get_setting( 'result_book' )
                book_file   = os.path.join( outp_folder, result_book )
                serial_file = serial_file = os.path.join( outp_folder + '/.temp', '--serial_result.csv' )
                guinier_folder = outp_folder + '/.guinier_result'
                stamp_file     = os.path.join( self.guinier_folder, '--stamp.csv' )
                controller = SerialController( data_folder, conc_folder, outp_folder,
                                                guinier_folder=guinier_folder,
                                                stamp_file=stamp_file,
                                                parent=self, serial_file=serial_file, book_file=book_file )

                data = self.conn.recv( BUFFER_SIZE )
                print( 'data=', data )
                # should be GET_NUM_STEPS

                num_steps  = controller.start()
                self.conn.send( bytes( str(num_steps), 'utf-8' ) )

                for i in range( num_steps ):
                    try:
                        data = self.conn.recv( BUFFER_SIZE )
                        if DEBUG: print( 'data=', data )
                        if data == REQUEST_INFO:
                            info = controller.get_info( block=True )
                            if DEBUG: print( 'info=', info )
                            self.conn.send( bytes( str(i+1), 'utf-8' ) )
                        elif data == CANCEL_PROCESS:
                            controller.kill()
                            self.conn.send( ACK )
                            break
                    except Exception as exc:
                        print( exc )
                        break
            else:
                print( 'Unexpected data=', data )
                self.conn.send( UNEXPECTED )
                break

        self.conn.close()
        if DEBUG: print( 'SerialControllerThread: terminated' )
        # any other thing to do for closing?

class AutorgKekClient:
    def __init__( self, host=HOST, port=PORT ):
        self.host   = host
        self.port   = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect( ( self.host, self.port ) )

    def __del__( self ):
        self.socket.close()

    def request_impl( self, *args ):
        ret = None
        for message in args:
            if DEBUG: print( "client sending message:", message )
            self.socket.send( message )
            ret = self.socket.recv( BUFFER_SIZE )
            if DEBUG: print( "client received data:", ret )
        return ret

    def stop_server( self ):
        self.request_impl( STOP )

    def request( self, *args ):
        # print( 'request_str: args=', *args )
        b_args = list( map( lambda x: x.encode() if type(x) == str else str( x ).encode(), args ) )
        ret = self.request_impl( *b_args )
        retcode = 0 if ret == SUCCESS else -1
        return retcode



"""
    SSDC.SsdcAnalysisUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCUV_PARAMS, compute_elution_curves

def plot_ssdc_result(names, info_list, rg_info_list, unreliable_indeces,
                     sdm_params_info_list, mapping_list, mapping,
                     extra_button_specs=[]):

    gs = GridSpec(2,4)
    with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
        fig = plt.figure(figsize=(20,8))
        axes_list = []
        for i in range(2):
            axes_row = []
            for j in range(3):
                jslice = j if j < 2 else slice(2,4)
                ax = fig.add_subplot(gs[i,jslice])
                axes_row.append(ax)
            axes_list.append(axes_row)
        axes = np.array(axes_list)

        fig.suptitle("SSDC - Same Sample in Different Columns Analysis: %s vs. %s" % tuple(names), fontsize=24)

        for k, (in_folder, lrfsrc) in enumerate(info_list):
            rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = rg_info_list[k]
            x = lrfsrc.xr_x
            y = lrfsrc.xr_y
            model = lrfsrc.model

            ax = axes[k,0]
            ax.set_title("EGH Decomposition of %s" % names[k], fontsize=16)
            ax.plot(x, y)
            cy_list = []
            peaks = lrfsrc.xr_peaks[indeces]
            for k, params in enumerate(peaks):
                cy = model(x, params)
                ax.plot(x, cy, ":", label="$R_g$=%.3g" % peak_rgs[k])
            ax.legend()
        
        for k, (in_folder, lrfsrc) in enumerate(info_list):
            x = lrfsrc.xr_x
            y = lrfsrc.xr_y
            ret_params, temp_rgs = sdm_params_info_list[k][0:2]
            cy_list, ty = compute_elution_curves(x, ret_params, temp_rgs)

            ax = axes[k,1]
            ax.set_title("SDM Decomposition of %s" % names[k], fontsize=16)
            ax.plot(x, y)
            for k, cy in enumerate(cy_list):
                ax.plot(x, cy, ":", label="$R_g$=%.3g" % temp_rgs[k])
            texts = ax.legend().get_texts()
            for i in unreliable_indeces:
                texts[i].set_color("red")

        slope, intercept = mapping

        for k, (in_folder, lrfsrc) in enumerate(info_list):
            x = lrfsrc.xr_x
            y = lrfsrc.xr_y
            ret_params, temp_rgs = sdm_params_info_list[k][0:2]
            cy_list, ty = compute_elution_curves(x, ret_params, temp_rgs)
            N, K, x0, poresize, N0, tI = ret_params[0:NUM_SDMCUV_PARAMS]

            ax = axes[k,2]
            ax.set_title("SDM Decomposition of %s with Poresize=%.3g, $N_0$=%.0f" % (names[k], poresize, N0), fontsize=16)
            ax.plot(x, y)
            moments = mapping_list[k]
            for j, cy in enumerate(cy_list):
                ax.plot(x, cy, ":", label="$R_g$=%.3g" % temp_rgs[j])
                ax.axvline(x=moments[j], color="green", alpha=0.5)
            ax.axvline(x=x0, color="red", label="$t_0$")
            ax.axvline(x=tI, color="gray", label="$t_{inj}$")
            # ax.set_xlim(-350, 1400)
            if k > 0:
                ax.set_xlim(*[x_*slope + intercept for x_ in [xmin, xmax]])
            xmin, xmax = ax.get_xlim()
            print([k], xmin, xmax)
            texts = ax.legend(bbox_to_anchor=(0.3, 1)).get_texts()
            for i in unreliable_indeces:
                texts[i].set_color("red")
            
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        plt.show()


# coding: utf-8
"""
    FactorAnalysis.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import matplotlib.pyplot    as plt
from sklearn.decomposition  import PCA, FastICA, FactorAnalysis as SKFA

DEBUG   = False

class FactorAnalysis:
    def __init__( self, a_curve, array ):

        num_factors = len( a_curve.peak_info )

        # fa = FastICA( n_components=num_factors, max_iter=5000 )
        fa = PCA( n_components=num_factors )
        X = fa.fit_transform( array )
        for i in range(num_factors):
            plt.plot(X[:,i])

        plt.show()



import numpy as np
import matplotlib.pyplot as plt

x = np.linspace(0, 1, 100)
y = np.linspace(0, 1, 100).reshape(-1,1)

fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))

data = x+y
print(x.shape, y.shape, data.shape)
ax1.imshow(data, animated=True)

fig.tight_layout()
plt.show()



# coding: utf-8
"""
    XrayData.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""

import numpy as np
from bisect import bisect_right
from MatrixData import MatrixData
from SerialDataUtils import load_intensity_files
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
from molass_legacy.Trimming.PreliminaryRg import get_default_angle_range_impl

class XrayData(MatrixData):
    def __init__(self, folder, array=None, data=None, q=None, error=None, sd=None, correct_elution=False):
        files = None
        if data is None:
            if array is None:
                if sd is None:
                    array, files = load_intensity_files(folder)
                    assert len(files) > 10      # temporary guard ot avoid confusing when given an invalid folder
                else:
                    array = sd.intensity_array
            q = array[0,:,0]
            data = array[:,:,1].T
            error = array[:,:,2].T
        else:
            assert q is not None
            assert error is not None

        self.vector = q
        MatrixData.__init__(self, data)
        self.error = MatrixData(error)
        self.set_elution_curve(correct=correct_elution)
        self.files = files

    def plot(self, ax=None, color=None, alpha=1, title="Xray Scattering", ec_color='orange'):
        MatrixData.plot(self, ax=ax, color=color, alpha=alpha, title=title, ec_color=ec_color)

    def set_elution_curve(self, correct=False):
        pick_pos = get_xray_picking()
        pick_num = get_setting('num_points_intensity')      # odd number
        index = bisect_right(self.vector, pick_pos)
        hwidth = pick_num//2
        # take max because (index - hwidth) can be negative, e.g. in 20190522/Backsub3_CytC
        start = max(0, index - hwidth)
        stop  = start + pick_num
        print('XrayData.set_elution_curve start, stop=', (start, stop))
        MatrixData.set_elution_curve(self, index, slice(start, stop), correct=correct)

    def set_restriction(self, mapping_params, flow_changes):
        angle_start, flange_limit, pre_rg = get_default_angle_range_impl(self.data, self.error.data, self.e_curve, self.vector, self.e_index, self.logger)
        print('angle_start, flange_limit=', angle_start, flange_limit)

        self.i_slice = slice(angle_start, flange_limit)

        slope, intercept = mapping_params
        A_, B_ = 1/slope, -intercept/slope
        mapped_fc = [None if j is None else min(self.data.shape[1], max(0, int(A_*j + B_ + 0.5))) for j in flow_changes]

        print('flow_changes=', flow_changes, 'mapped_fc=', mapped_fc)
        if mapped_fc[0] is None:
            mapped_fc[0] = 0
        self.j_slice = slice(*mapped_fc)

    def copy_for_mct(self):
        i_slice = self.i_slice
        data = self.data[i_slice,:]
        q = self.vector[i_slice]
        error = self.error.data[i_slice,:]
        copy = XrayData(None, data=data, q=q, error=error)
        # copy.plot()
        return copy



## Generate a contour plot
# Import some other libraries that we'll need
# matplotlib and numpy packages must also be installed
import matplotlib
import numpy as np
import matplotlib.pyplot as plt
import random
import math

# define objective function
def f(x):
    x1 = x[0]
    x2 = x[1]
    obj = 0.2 + x1**2 + x2**2 - 0.1*math.cos(6.0*3.1415*x1) - 0.1*math.cos(6.0*3.1415*x2)
    return obj

# Start location
x_start = [0.8, -0.5]

# Design variables at mesh points
i1 = np.arange(-1.0, 1.0, 0.01)
i2 = np.arange(-1.0, 1.0, 0.01)
x1m, x2m = np.meshgrid(i1, i2)
fm = np.zeros(x1m.shape)
for i in range(x1m.shape[0]):
    for j in range(x1m.shape[1]):
        fm[i][j] = 0.2 + x1m[i][j]**2 + x2m[i][j]**2 \
             - 0.1*math.cos(6.0*3.1415*x1m[i][j]) \
             - 0.1*math.cos(6.0*3.1415*x2m[i][j])

# Create a contour plot
plt.figure()
# Specify contour lines
#lines = range(2,52,2)
# Plot contours
CS = plt.contour(x1m, x2m, fm)#,lines)
# Label contours
plt.clabel(CS, inline=1, fontsize=10)
# Add some text to the plot
plt.title('Non-Convex Function')
plt.xlabel('x1')
plt.ylabel('x2')

##################################################
# Simulated Annealing
##################################################
# Number of cycles
n = 50
# Number of trials per cycle
m = 50
# Number of accepted solutions
na = 0.0
# Probability of accepting worse solution at the start
p1 = 0.7
# Probability of accepting worse solution at the end
p50 = 0.001
# Initial temperature
t1 = -1.0/math.log(p1)
# Final temperature
t50 = -1.0/math.log(p50)
# Fractional reduction every cycle
frac = (t50/t1)**(1.0/(n-1.0))
# Initialize x
x = np.zeros((n+1,2))
x[0] = x_start
xi = np.zeros(2)
xi = x_start
na = na + 1.0
# Current best results so far
xc = np.zeros(2)
xc = x[0]
fc = f(xi)
fs = np.zeros(n+1)
fs[0] = fc
# Current temperature
t = t1
# DeltaE Average
DeltaE_avg = 0.0
for i in range(n):
    print('Cycle: ' + str(i) + ' with Temperature: ' + str(t))
    for j in range(m):
        # Generate new trial points
        xi[0] = xc[0] + random.random() - 0.5
        xi[1] = xc[1] + random.random() - 0.5
        # Clip to upper and lower bounds
        xi[0] = max(min(xi[0],1.0),-1.0)
        xi[1] = max(min(xi[1],1.0),-1.0)
        DeltaE = abs(f(xi)-fc)
        if (f(xi)>fc):
            # Initialize DeltaE_avg if a worse solution was found
            #   on the first iteration
            if (i==0 and j==0): DeltaE_avg = DeltaE
            # objective function is worse
            # generate probability of acceptance
            p = math.exp(-DeltaE/(DeltaE_avg * t))
            # determine whether to accept worse point
            if (random.random()<p):
                # accept the worse solution
                accept = True
            else:
                # don't accept the worse solution
                accept = False
        else:
            # objective function is lower, automatically accept
            accept = True
        if accept:
            # update currently accepted solution
            xc[0] = xi[0]
            xc[1] = xi[1]
            fc = f(xc)
            # increment number of accepted solutions
            na = na + 1.0
            # update DeltaE_avg
            DeltaE_avg = (DeltaE_avg * (na-1.0) +  DeltaE) / na
    # Record the best x values at the end of every cycle
    x[i+1][0] = xc[0]
    x[i+1][1] = xc[1]
    fs[i+1] = fc
    # Lower the temperature for next cycle
    t = frac * t

# print solution
print('Best solution: ' + str(xc))
print('Best objective: ' + str(fc))

plt.plot(x[:,0],x[:,1],'y-o')
# plt.savefig('contour.png')

fig = plt.figure()
ax1 = fig.add_subplot(211)
ax1.plot(fs,'r.-')
ax1.legend(['Objective'])
ax2 = fig.add_subplot(212)
ax2.plot(x[:,0],'b.-')
ax2.plot(x[:,1],'g--')
ax2.legend(['x1','x2'])

# Save the figure as a PNG
# plt.savefig('iterations.png')

plt.show()



# coding: utf-8
"""
    ProbDataUtils.py.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit
import molass_legacy.KekLib.DebugPlot as plt

@jit(nopython=True)
def generate_sample_data_impl(y, p, ymax):
    n = 10**p
    data = []
    for i, y_ in enumerate(y):
        for j in range(int(y_/ymax*n)):
            data.append(i)
    return np.array(data)

def generate_sample_data(y, p):
    ymax = np.max(y)
    return generate_sample_data_impl(y, p, ymax)

def set_consistent_base(ax1, axt):
    ymin1, ymax1 = ax1.get_ylim()
    ymint, ymaxt = axt.get_ylim()
    y_ = ymin1/ymax1*ymaxt
    axt.set_ylim(y_, ymaxt)

def plot_hist_data(y1, y2, data1, data2, axes=None, plot_class=plt):
    from DataUtils import get_in_folder

    if axes is None:
        fig, axes = plot_class.subplots(nrows=1, ncols=2, figsize=(14,7))
        fig.suptitle("Histogram of Generated Data overlaid with the Elution Curve from %s" % get_in_folder(), fontsize=20)

    ax1, ax2 = axes
    ax1t = ax1.twinx()
    ax2t = ax2.twinx()
    ax1t.grid(False)
    ax2t.grid(False)

    ax1.set_title("UV Data", fontsize=16)
    ax2.set_title("X-ray Data", fontsize=16)

    ax1.plot(y1)
    ax2.plot(y2)
    ax1t.hist(data1, bins=len(y1), alpha=0.5, color='orange')
    set_consistent_base(ax1, ax1t)
    ax2t.hist(data2, bins=len(y2), alpha=0.5, color='orange')
    set_consistent_base(ax2, ax2t)

    if axes is None:
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)

def generate_samle_datasets(in_folder, kwargs):
    lpm_correct = kwargs.pop('lpm_correct', False)
    lpm_2d = kwargs.pop('lpm_2d', False)
    smoothing = kwargs.get('smoothing', False)

    if lpm_correct:
        from CorrectedData import  CorrectedXray,  CorrectedUv
        ru = CorrectedUv(in_folder)
        rx = CorrectedXray(in_folder)
    else:
        from RawData import RawXray, RawUv
        ru = RawUv(in_folder)
        rx = RawXray(in_folder)

    i = ru.get_row_index(280)
    y1 = ru.data[i,:]
    if lpm_2d:
        from LPM import get_corrected
        y1 = get_corrected(y1)

    i = rx.get_row_index(0.02)
    y2 = rx.data[i,:]
    if lpm_2d:
        from LPM import get_corrected
        y2 = get_corrected(y2)

    if smoothing:
        from molass_legacy.KekLib.SciPyCookbook import smooth
        y1 = smooth(y1)
        y2 = smooth(y2)

    data1 = generate_sample_data(y1, 2)
    data2 = generate_sample_data(y2, 2)

    return y1, y2, data1, data2



# coding: utf-8
"""

    ファイル名：   MachineTypes.py

    処理内容：

        マシンタイプ（Chassis Type）などの識別

    Copyright (c) 2015-2019, Masatsuyo Takahashi, KEK-PF

"""

import wmi

CHASSIS_TYPE_NAMES = {
    1   :   "Other",
    2   :   "Unknown",
    3   :   "Desktop",
    4   :   "Low Profile Desktop",
    5   :   "Pizza Box",
    6   :   "Mini Tower",
    7   :   "Tower",
    8   :   "Portable",
    9   :   "Laptop",
    10  :   "Notebook",
    11  :   "Hand Held",
    12  :   "Docking Station",
    13  :   "All in One",
    14  :   "Sub Notebook",
    15  :   "Space-Saving",
    16  :   "Lunch Box",
    17  :   "Main System Chassis",
    18  :   "Expansion Chassis",
    19  :   "SubChassis",
    20  :   "Bus Expansion Chassis",
    21  :   "Peripheral Chassis",
    22  :   "Storage Chassis",
    23  :   "Rack Mount Chassis",
    24  :   "Sealed-Case PC",
    }

def get_chassistype_name():
    wmi_ = wmi.WMI()
    type = None
    for enclosure in wmi_.Win32_SystemEnclosure():
        type = enclosure.ChassisTypes[0]
        break;
    assert( type != None )
    return CHASSIS_TYPE_NAMES[ type ]

def get_display_resolution():
    wmi_ = wmi.WMI()
    width, height = None, None
    for config in wmi_.Win32_DisplayConfiguration():
        width   = config.PelsWidth
        height  = config.PelsHeight
        # print( width, height )
        break
    return ( width, height )

monitors = []

def get_monitors():
    global monitors

    if len( monitors ) > 0: return monitors

    wmi_ = wmi.WMI()
    # for info in wmi_.Win32_DisplayConfiguration():
    # for info in wmi_.Win32_DisplayControllerConfiguration():
    # for info in wmi_.Win32_VideoController():
    # for info in wmi_.Win32_VideoConfiguration():
    # for info in wmi_.Win32_VideoSettings():
    # for info in wmi_.Win32_DesktopMonitor():
    monitors = []
    for info in wmi_.Win32_PnPEntity():
        if info.Service == 'monitor':
            # TODO: set where condition to the above call.
            # print( info.HardwareID )
            monitors.append( info.HardwareID[0].split( '\\' )[1]  )

    return monitors

def get_Win32_Processor_str():
    wmi_ = wmi.WMI()
    return str(wmi_.Win32_Processor()[0])

def is_on_virtual_machine():
    s = get_Win32_Processor_str()
    return s.find('ProcessorId') < 0

def get_cpuid():
    import re
    s = get_Win32_Processor_str()
    cpuid_re = re.compile(r'ProcessorId = "(\w+)"')
    m = cpuid_re.search(s)
    return m.group(1) if m else None

def get_Win32_ComputerSystem_str():
    wmi_ = wmi.WMI()
    return str(wmi_.Win32_ComputerSystem()[0])

def get_bootup_state():
    import re
    s = get_Win32_ComputerSystem_str()
    # print(s)
    state_re = re.compile(r'BootupState = "(.+)";')
    m = state_re.search(s)
    if m:
        return m.group(1)



"""
    StcParams.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from .MonoporeSecParams import MonoporeSecParams, NUM_SEC_PARAMS, sec_comformance_impl
from .BaselineParams import get_num_baseparams
from molass_legacy._MOLASS.SerialSettings import get_setting

def get_common_parameter_names(nc):
    xr_names = ["$h_%d$" % k for k in range(nc)]
    rg_names = ["$R_{g%d}$" % k for k in range(nc)]
    mapping_names = ["$mp_a$", "$mp_b$"]
    uv_names = ["$uh_%d$" % k for k in range(nc)]
    mr_names = ["$mr_a$", "$mr_b$"]
    seccol_names = ["$t_0$", "$r_p$", "$N$", "$me$", "$T$", "mp"]

    return xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names

class StcParamsBase:
    def __init__(self, n_components, poresize, poreexponent, use_K=False):
        self.n_components = n_components
        self.num_baseparams = get_num_baseparams()
        self.integral_baseline = self.num_baseparams == 3
        self.poresize = poresize
        if poreexponent == 0:
            # workaround for set_setting("poreexponent", None) not working
            poreexponent = None
        self.poreexponent = poreexponent
        sectype = MonoporeSecParams(poresize, poreexponent)

        """
        t0_upper_bound will be set as follows
            1.  initialzied from get_setting("t0_upper_bound")
                which is None as first in the monitor process,
                but has already been given a proper value in the optimizer process

            2.  in the the monitor process, it will be given a proper value later
                at MonoporeComponentParams.get_estimator,
                and then, it will be passed to the optimizer process
        """
        self.t0_upper_bound = sectype.t0_upper_bound

        self.init_method = sectype.init_method
        # self.conf_method = sectype.conf_method
        self.estm_method = sectype.estm_method
        self.nump_adjust = sectype.nump_adjust
        self.use_K = use_K

    def compute_init_guess(self, *args):
        return self.init_method(*args)

    def compute_comformance(self, *args, **kwargs):
        # return self.conf_method(*args, **kwargs)      # caused a bug in EGH model in Python 3.10.7
        return sec_comformance_impl(self.t0_upper_bound, *args, **kwargs)   # may be safer

    def estimate_conformance_params(self, *args):
        return self.estm_method(*args)

    def get_param_bounds(self, params):

        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_params, init_uv_baseparams, init_mappable_range = self.split_params_simple(params)[0:7]

        xr_h_max = np.max(init_xr_params[:])
        m_allow = 100

        xr_bounds = []
        for h in init_xr_params:
            xr_bounds.append((0, xr_h_max*2))       # h

        for k, v in enumerate(init_xr_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 2:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            xr_bounds.append(bounds)

        rg_bounds = [(rg*0.5, rg*2) for rg in init_rgs]
        a, b = init_mapping
        mapping_bounds = [(a*0.8, a*1.2), (-m_allow, m_allow)]

        uv_h_max = np.max(init_uv_params)
        uv_bounds = [(0, uv_h_max*2) for h in init_uv_params]
        for k, v in enumerate(init_uv_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 7:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            uv_bounds.append(bounds)

        f, t = init_mappable_range
        dx = (t - f)*0.1
        range_bounds = [(f-dx, f+dx), (t-dx, t+dx)]

        # t0, rp, N, me, T, mp
        if self.use_K:
            # K = N*T
            # m = me + mp
            # p = log(T)/log(K)
            # q = mp/m
            sec_bounds = [(-100, 100), (30, 150), (200, 3000), (1, 3), (-0.1, 1.1), (-0.2, 1.2)]
        else:
            sec_bounds = [(-100, 100), (30, 150), (100, 2000), (1, 3), (0.1, 10), (1, 3)]
        self.bounds_lengths = [len(b) for b in [xr_bounds, rg_bounds, mapping_bounds, uv_bounds, range_bounds, sec_bounds]]
        return xr_bounds + rg_bounds + mapping_bounds + uv_bounds + range_bounds + sec_bounds

    def split_bounds(self, bounds):
        separate_bounds = []
        offset = 0
        for k, length in enumerate(self.bounds_lengths):
            next_offset = offset + length
            sep_bounds = bounds[offset:next_offset]
            if k == 0:
                sep_bounds = self.reshape_xr_bounds(sep_bounds)
            separate_bounds.append(sep_bounds)
            offset = next_offset
        return separate_bounds

    def update_bounds_hook(self, masked_init_params):
        # nothing to do
        pass

    def get_peak_pos_array_list(self, x_array):
        nc = self.n_components - 1
        gr_start = self.get_rg_start_index()

        pos_array_list = []
        for params in x_array:
            rg_params = params[gr_start:gr_start+nc]
            t0, rp, N, me, T, mp = params[-NUM_SEC_PARAMS:]
            rho = rg_params/rp
            rho[rho > 1] = 1
            model_trs = t0 + N*T*(1 - rho)**(me + mp)
            pos_array_list.append(model_trs)

        return np.array(pos_array_list).T

    def get_params_sheet(self, parent, params, dsets, optimizer, debug=True):
        if debug:
            from importlib import reload
            import ModelParams.StcParamsSheet
            reload(ModelParams.StcParamsSheet)
        from .StcParamsSheet import StcParamsSheet
        return StcParamsSheet(parent, params, dsets, optimizer)

    def get_adjuster(self, debug=True):
        if debug:
            import ModelParams.StcAdjuster
            from importlib import reload
            reload(ModelParams.StcAdjuster)
        from .StcAdjuster import StcAdjuster
        return StcAdjuster()

class MonoporeComponentParams(StcParamsBase):
    def __init__(self, n_components, poresize=None, poreexponent=None, use_K=False):
        StcParamsBase.__init__(self, n_components, poresize, poreexponent, use_K=use_K)
        self.advanced = True
        nc = n_components - 1

        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = nc
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params


    def __str__(self):
        ps, px = [str(v) if v is None else "%g" % v for v in [self.poresize, self.poreexponent]]
        return "MonoporeComponentParams(nc=%d, ps=%s, px=%s)" % (self.n_components, ps, px)

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import Estimators.StcEstimator
            reload(Estimators.StcEstimator)
        from Estimators.StcEstimator import StcEstimator

        estimator = StcEstimator(editor)
        self.t0_upper_bound = estimator.get_t0_upper_bound()
        return estimator

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        ret_params = []
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            ret_params.append(params[p:q])

        return ret_params

    def split_params_simple(self, params):
        r = len(params) - self.num_params
        if r == 0:
            decomp_params = params
            seccol_params = None
        else:
            assert r == NUM_SEC_PARAMS
            decomp_params = params[:-NUM_SEC_PARAMS]
            seccol_params = params[-NUM_SEC_PARAMS:]

        self.separate_params = self.split_params(self.n_components, decomp_params) + [seccol_params]
        return self.separate_params

    def split_as_unified_params(self, params):
        return self.split_params_simple(params) + [(None, None)]

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + NUM_SEC_PARAMS, dtype=bool)
        n = self.n_components
        nc = n - 1
        bounds_mask[0:nc] = True            # xr_params
        sep = nc + self.num_baseparams

        if self.integral_baseline:
            bounds_mask[sep-1] = True           # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv aseline fouling

        bounds_mask[sep:sep+nc] = True      # rgs

        return bounds_mask

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def get_parameter_names(self):
        xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(self.n_components - 1)
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_trans_indeces(self):
        return -4, -3, -2, -1

    def split_get_unified_sec_params(self, params):
        t0, rp, N, me, T, mp = self.split_params_simple(params)[-1]
        # K = N*T
        # m = me + mp
        return t0, N*T, rp, me+mp, N, T, me, mp, None, None, None

def construct_stochastic_params_type(n_components, use_K=False):
    poresize = get_setting("poresize")
    poreexponent = get_setting("poreexponent")
    uv_basemodel = get_setting("uv_basemodel")
    if uv_basemodel == 1:
        params_type = MonoporeComponentParams(n_components, poresize, poreexponent, use_K=use_K)
    else:
        assert False
    return params_type



"""
    Models.Stochastic.MonoporeMoments.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize, basinhopping
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.Models.Stochastic.MomentUtils import to_moderate_props
from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, M1_WEIGHT, M2_WEIGHT, BASINHOPPING_SCALE, MonoporeParamsScaler

def collate_monopore_moments():
    print("collate_monopore_moments")

def study_monopore_moments_impl(lrf_src,
                                use_basinhopping = True,        # it seems that global optimization is required for this problem
                                use_timescale = True,           # timescale parameter, which will be removed eventually,
                                                                # is used to help the optimizer by relating the changes of t0 and T
                                keep_num_components=False,
                                want_num_components=4,          # this should be eventually want_num_components=None
                                trust_all_rgs=False,            # used for illustration
                                return_init_params=False,       # uses for animation
                                return_init_moments=False,      # used for illustration
                                return_step1_moments=False,     # used for illustration
                                return_step2_moments=False,     # used for illustration
                                trust_max_num=None,
                                debug=False):
    if debug:
        import Models.Stochastic.MonoporeUtils as module
        reload(module)
        import Models.Stochastic.ExclusionCurves as module
        reload(module)

    from molass_legacy.Models.Stochastic.MonoporeUtils import guess_monopore_colparams
    from molass_legacy.Models.Stochastic.ExclusionCurves import compute_simply_conformant_rgs

    print('study_monopore_moments_impl')
    logger = lrf_src.logger

    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(keep_num_components=keep_num_components,
                                                                                              want_num_components=want_num_components,  debug=False)
    logger.info("study_monopore_moments_impl entry: peak_rgs=%s", peak_rgs)
    peaks = lrf_src.get_peaks()
    num_peaks = len(peaks)
    egh_moments_list = lrf_src.get_egh_moments_list()
    m_props = to_moderate_props(props)

    me = 1.5
    mp = 1.5

    if trust_all_rgs:
        unreliable_indeces = np.array([], dtype=int)
    else:
        unreliable_indeces = determine_unreliables(peak_rgs, qualities, props, trust_max_num=trust_max_num)
    nur = len(unreliable_indeces)

    use_conformant_rgs = False

    adjust_rgs_at_start_only = False
    u_indeces = np.array([], dtype=int) if (use_conformant_rgs or adjust_rgs_at_start_only) else unreliable_indeces
    num_timescales = 1 if use_timescale else 0
    params_scaler = MonoporeParamsScaler(egh_moments_list, 0, peak_rgs, u_indeces, num_timescales=num_timescales, allow_near_t0=False)

    temp_rgs = np.asarray(peak_rgs).copy()
    logger.info("initial input info: temp_rgs=%s, unreliable_indeces=%s", temp_rgs, unreliable_indeces)
    egh_moments_array = np.asarray(egh_moments_list)
    if num_peaks > 1:
        eghm1_min_diff = np.min(np.abs(np.diff(egh_moments_array[:,0])))
        if debug:
            print("eghm1_min_diff=", eghm1_min_diff)
        MIN_DIFF_PENALTY_SCALE = 1e6
    
    apply_diff_penalty = True
    def mnp_column_flex_objective(p, return_moments=False, debug=False):
        if use_basinhopping:
            p = params_scaler.scale_back(p)
        N_, KT_, x0_, poresize_ = p[0:4]
        if USE_K:
            T_ = KT_/N_
        else:
            T_ = KT_
        if use_timescale:
            T_ *= p[-1]
            x0_ *= p[-1]
        if not adjust_rgs_at_start_only:
            if nur > 0:
                if use_conformant_rgs:
                    temp_rgs[unreliable_indeces] = compute_simply_conformant_rgs(N_, KT_, x0_, poresize_, 3, peak_trs[unreliable_indeces], debug=debug, rgs=peak_rgs[unreliable_indeces])
                else:
                    temp_rgs[unreliable_indeces] = p[4:4+nur]

        if len(temp_rgs) > 1:
            # Rg's should be in descending order
            order_penalty = max(0, np.max(np.diff(temp_rgs))) * MIN_DIFF_PENALTY_SCALE
        else:
            order_penalty = 0

        rhov = temp_rgs/poresize_
        rhov[rhov > 1] = 1
        dev_list = []
        moments_list = []
        for k, (M, rho) in enumerate(zip(egh_moments_list, rhov)):
            M1_ = x0_ + N_ * T_ * (1 - rho)**(me + mp)
            M2_ = 2 * N_ * T_**2 * (1 - rho)**(me + 2*mp)
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if return_moments:
                moments_list.append([M1_, M2_])
            else:
                moments_list.append(M1_)
            if debug:
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))

        if return_moments:
            return moments_list

        if apply_diff_penalty and num_peaks > 1:
            m1_min_diff = np.min(np.abs(np.diff(moments_list)))
            diff_penalty = max(0, abs(1 - m1_min_diff/eghm1_min_diff) - 0.1)**2 * MIN_DIFF_PENALTY_SCALE
        else:
            diff_penalty = 0
        fv = np.sum(np.asarray(dev_list) * m_props) + diff_penalty + order_penalty   # seems better not to weight with props
        if debug:
            print("m_props=", m_props)
            print("order_penalty=", order_penalty)
            print("diff_penalty=", diff_penalty)
            print("fv=", fv)
        return fv

    if USE_K:
        kt_init = np.sqrt(np.sum(egh_moments_array[:,1])) * 30
        # kt_init = 500
    else:
        kt_init = 0.5
    t0_init = np.min([M[0] - np.sqrt(M[1])*20 for M in egh_moments_list])
    colparams = [1000, kt_init, t0_init, 200]

    plot_info = Struct(x=lrf_src.xr_x, y=lrf_src.xr_y, model=lrf_src.model, peaks=peaks)
    if num_peaks > 3:
        ret = guess_monopore_colparams(peak_rgs, qualities, peak_trs, colparams, debug=debug, plot_info=plot_info)
        if ret is None:
            return
        colparams = list(ret)
        print("colparams=", colparams)

    rgs = peak_rgs[unreliable_indeces]
    if use_conformant_rgs:
        conformant_rgs = []
    else:
        trs = peak_trs[unreliable_indeces]
        ret = compute_simply_conformant_rgs(*colparams, 3, trs, debug=debug, rgs=rgs, plot_info=plot_info)
        if ret is None:
            return
        temp_rgs[unreliable_indeces] = ret
        conformant_rgs = list(ret)
        logger.info("using conformant Rg's %s from trs=%s, for unreliable peaks at start only", conformant_rgs, trs)

    timescale = [1] if use_timescale else []
    param_rgs = [] if adjust_rgs_at_start_only else conformant_rgs
    init_params = colparams + param_rgs + timescale
 
    if return_init_params:
        return init_params

    if use_basinhopping:
        init_params_ = params_scaler.scale(init_params)
        if return_init_moments:
            return mnp_column_flex_objective(init_params_, return_moments=True)

        bounds_ = [(0, BASINHOPPING_SCALE)] * len(init_params)
        if debug:
            print("mnp_column_objective:init_params=", init_params)
            print("mnp_column_objective:init_params_=", init_params_)
            print("mnp_column_objective:bounds=", params_scaler.get_bounds())
            print("mnp_column_objective:bounds_=", bounds_)
            mnp_column_flex_objective(init_params_, debug=True)

        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds_)
        # minimizer_kwargs = dict(method='trust-constr', bounds=bounds_)
        # minimizer_kwargs = dict(method='SLSQP', bounds=bounds_)
        # minimizer_kwargs = dict(method='COBYLA', bounds=bounds_)
        # minimizer_kwargs = dict(bounds=bounds_)
        res = basinhopping(mnp_column_flex_objective, init_params_,
                           # stepsize=0.5,
                           niter=100,
                           minimizer_kwargs=minimizer_kwargs)

        mnp_column_flex_objective(res.x, debug=True)
        if return_step1_moments:
            return mnp_column_flex_objective(res.x, return_moments=True)

        res.x = params_scaler.scale_back(res.x)     # do not forget to scale back the result
    else:
        bounds = params_scaler.get_bounds()
        if debug:
            print("mnp_column_objective:init_params=", init_params)
            print("mnp_column_objective:bounds=", bounds)
        
        res = minimize(mnp_column_flex_objective, init_params, bounds=bounds, method='Nelder-Mead')
        if return_step1_moments:
            return mnp_column_flex_objective(res.x, return_moments=True)

    if debug:
        print("mnp_column_objective: res.x=", res.x)
    
    x = lrf_src.xr_x
    y = lrf_src.xr_y

    N, KT, x0, poresize = res.x[0:4]
    if USE_K:
        T = KT/N
    else:
        T = KT
    if use_timescale:
        T *= res.x[-1]
        x0 *= res.x[-1]
    if nur > 0:
        if use_conformant_rgs:
            temp_rgs[unreliable_indeces] = compute_simply_conformant_rgs(N, KT, x0, poresize, 3, peak_trs[unreliable_indeces])
        else:
            temp_rgs[unreliable_indeces] = res.x[4:4+nur]
    rhov = temp_rgs/poresize
    rhov[rhov > 1] = 1
    abort = False

    def mnp_scales_objective(p, return_cy_list=False, return_moments=False, title=None):
        debug = title is not None
        T_ = T*p[-1]
        x0_ = p[0] * p[-1]
        scales = p[1:1+num_peaks]
        cy_list = []
        if debug or return_moments:
            moments_list = []
        for k, (rho, scale) in enumerate(zip(rhov, scales)):
            ni_ = N * (1 - rho)**me
            ti_ = T_ * (1 - rho)**mp
            if debug or return_moments:
                M1_ = x0_ + ni_ * ti_
                M2_ = 2 * ni_ * ti_**2
                moments_list.append([M1_, M2_])
            cy = scale * robust_single_pore_pdf(x - x0_, ni_, ti_)
            cy_list.append(cy)
        if return_cy_list:
            return cy_list
        if return_moments:
            return moments_list
        ty = np.sum(cy_list, axis=0)
        dev = np.sum((ty - y)**2)
        if debug:
            from molass_legacy.Models.Stochastic.MonoporeUtils import plot_monopore_moments_state
            title += "; FV=%.3g" % dev
            mnp_params = np.concatenate([[N, T, x0_, me, mp, poresize], scales])
            bounds = params_scaler.get_bounds()
            t0_upper = bounds[2][1]
            ret = plot_monopore_moments_state(title, x, y, temp_rgs, mnp_params, t0_upper, cy_list, ty, moments_list, egh_moments_list)
            if not ret:
                nonlocal abort
                abort = True
        return dev
    
    x0_ones = np.concatenate([[x0], np.ones(num_peaks)])
    cy_list = mnp_scales_objective(x0_ones, return_cy_list=True)
    init_scales = []
    for cy, h in zip(cy_list, peaks[:,0]):
        init_scales.append(h/np.max(cy))
    timescale = [1] if use_timescale else []
    init_params = np.concatenate([[x0], init_scales, timescale])
    if debug:
        mnp_scales_objective(init_params, title="before MNP scales optimize with unreliables %s minimize" % unreliable_indeces)
        if abort:
            return
    bounds = [(-1000, 1000)] + [(0, 10)]*num_peaks + [(0, 2)]
    print("init_params=", init_params)
    print("bounds=", bounds)
    res = minimize(mnp_scales_objective, init_params, bounds=bounds, method='Nelder-Mead')
    if debug:
        mnp_scales_objective(res.x, title="after MNP scales optimize with unreliables %s minimize" % unreliable_indeces)
        if abort:
            return

    if return_step2_moments:
        return mnp_scales_objective(res.x, return_moments=True), unreliable_indeces

    x0 = res.x[0]
    if use_timescale:
        x0 *= res.x[-1]
        T *= res.x[-1]
    scales = res.x[1:1+num_peaks]
    mnp_params = np.concatenate([(N, T, x0, me, mp, poresize), scales])
    if debug and use_timescale:
        temp_params = np.concatenate([[x0], scales, timescale])
        mnp_scales_objective(temp_params, title="timescale proof")
        if abort:
            return
    return mnp_params, temp_rgs, unreliable_indeces, params_scaler


"""
    RobustPeaks.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
from scipy.stats import linregress
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
import molass_legacy.KekLib.DebugPlot as plt

COUNT_FOR_BREAK = 3
MIN_HEIGHT_RATIO = 0.05
EXTEND_HEIGHT_RATIO = 0.02
EXTREME_POS_RATIO = 0.95        # < 0.99 for 20201007_2

def guess_peak_width(x, y, pt, min_width, debug=False):
    from scipy.stats import linregress

    r_values = []
    last_r_value = None
    dec_count = 0
    for i in range(min_width, pt-min_width):
        x_ = x[pt-i:pt]
        y_ = y[pt-i:pt]
        slope, intercept, r_value, p_value, std_err = linregress(x_, y_)
        r_values.append(r_value)
        if last_r_value is not None and r_value < last_r_value:
            dec_count += 1
            if dec_count > COUNT_FOR_BREAK:
                # i.e., more than COUNT_FOR_BREAK consecutive decrease
                break
        else:
            dec_count = 0
        last_r_value = r_value

    left_width = min_width + np.argmax(r_values)

    if debug:
        plt.push()
        fig, ax = plt.subplots()
        ax.set_title("Left of %d" % pt)
        ax.plot(x, y)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        p = pt - left_width
        ax.plot([p, p], [ymin, ymax], ':', color='red')
        axt = ax.twinx()
        axt.grid(False)
        f = pt - min_width - len(r_values)
        t = f + len(r_values)
        axt.plot(x[f:t], list(reversed(r_values)), color='C1')
        plt.show()
        plt.pop()

    r_values = []
    last_r_value = None
    inc_count = 0
    for i in range(min_width, len(x)-pt-min_width):
        x_ = x[pt:pt+i]
        y_ = y[pt:pt+i]
        slope, intercept, r_value, p_value, std_err = linregress(x_, y_)
        r_values.append(r_value)
        if last_r_value is not None and r_value > last_r_value:
            inc_count += 1
            if inc_count > COUNT_FOR_BREAK:
                # i.e., more than COUNT_FOR_BREAK consecutive increase
                break
        else:
            inc_count = 0
        last_r_value = r_value

    right_width = min_width + np.argmin(r_values)

    if debug:
        plt.push()
        fig, ax = plt.subplots()
        ax.set_title("Right of %d" % pt)
        ax.plot(x, y)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        p = pt + right_width
        ax.plot([p, p], [ymin, ymax], ':', color='red')
        axt = ax.twinx()
        axt.grid(False)
        f = pt + min_width
        t = f + len(r_values)
        axt.plot(x[f:t], r_values, color='C1')
        plt.show()
        plt.pop()

    return left_width, right_width

def fix_bubble(pt, y):
    ny = y[pt-5:pt+5]
    h = ny[-1] - ny[0]
    by = ny[0] + np.arange(10)*h/9
    ny_ = ny - by
    pp = np.where(ny_ > abs(h)/2)[0]
    pp_ = pt-5+pp
    y[pp_] -= ny_[pp]

    if False:
        print("abs(h)/2=", abs(h)/2)
        print("pp=", pp)
        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
        ax1.plot(ny_)
        ax2.plot(y)
        ymin, ymax = ax2.get_ylim()
        for p in pp_[[0,-1]]:
            ax2.plot([p, p], [ymin, ymax], ':', color='red')
        fig.tight_layout()
        plt.show()
        plt.pop()

    return pp_[[0,-1]]

def recognize_peaks_roughly(x, y, logger=None, num_peaks=None, min_ratio=MIN_HEIGHT_RATIO, min_width=10, check_bubble=False, return_ignore_info=False, debug=False):
    # min_width=10 is too small for 20200630_8

    y_copy = y.copy()
    pt = np.argmax(y_copy)
    pos_ratio = pt/len(y_copy)
    ignore_info = None
    if pos_ratio > EXTREME_POS_RATIO:
        from molass_legacy.KekLib.BasicUtils import Struct
        from GeometryUtils import rotated_argmin
        # like ths last peak in in 20201007_2
        logger.warning("an abnormal peak will be ignored due to pos_ratio=%.3g > %.3g", pos_ratio, EXTREME_POS_RATIO)
        j = rotated_argmin(-np.pi/4, y_copy, debug=False)
        y_copy[j:] = 0
        pt = np.argmax(y_copy)
        ignore_info = Struct(slice_=slice(j, None))

    y_max = y_copy[pt]

    half_min_width = min_width//2
    if check_bubble:
        bubble_check_ratio = np.average(y[pt-half_min_width:pt+half_min_width])/y_max
        print("bubble_check_ratio=", bubble_check_ratio)
        if bubble_check_ratio < 0.5:
            # for bubles in e.g. 20170226/Sugiyama
            f, t = fix_bubble(pt, y_copy)
            pt_ = np.argmax(y_copy)
            y_max = y_copy[pt]
            logger.warning("suspected bubble at %d:%d has been fixed to get a new peak top %d.", f, t+1, pt_)
            pt = pt_

    peak_tops = []
    peaks = []
    for k in range(10):
        try:
            left_width, right_width = guess_peak_width(x, y_copy, pt, min_width, debug=debug)
        except:
            log_exception(logger, "guess_peak_width failed to end looking for peaks: ")
            break

        print([k], "left_width, right_width=", left_width, right_width)
        if min(left_width, right_width) < min_width:
            break

        height_ratio = y_copy[pt]/min_ratio
        local_height_ratio = (y_copy[pt] - np.min(y_copy[[pt-left_width, pt+right_width]]))/y_max
        if local_height_ratio < min_ratio and height_ratio < min_ratio*2:
            break

        if debug:
            def debug_plot(title):
                print("left_width, right_width=", left_width, right_width)
                plt.push()
                fig, ax = plt.subplots()
                ax.set_title(title + "[%d]-th debug plot" % k)
                ax.plot(y)
                ax.plot(y_copy)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for w in [-left_width*2, -left_width, right_width, right_width*2]:
                    b = pt + w
                    ax.plot([b, b], [ymin, ymax], ':', color='yellow')
                fig.tight_layout()
                plt.show()
                plt.pop()
            debug_plot("before crush ")

        i = bisect_right(peak_tops, pt)
        peak_tops.insert(i, pt)
        print([k], "peak_tops=", peak_tops)
        peaks.insert(i, [pt-left_width, pt, pt+right_width])
        if num_peaks is not None and len(peaks) == num_peaks:
            break

        f_, t_ = get_covered_range(x, y_copy, peaks, i, pt, left_width, right_width, logger)
        y_copy[f_:t_] = 0
        pt_next = np.argmax(y_copy)

        if debug:
            print("pt_next=", pt_next)
            print("y_copy[pt_next]/y_max=", y_copy[pt_next]/y_max, min_ratio)
            debug_plot("after crush ")

        if len(np.where(y_copy[pt_next-half_min_width:pt_next+half_min_width] == 0)[0])/min_width > 0.3:
            break

        if y_copy[pt_next]/y_max < min_ratio:
            break

        pt = pt_next

    if return_ignore_info:
        return peaks, ignore_info
    else:
        return peaks

def get_covered_range(x, y, peaks, i, pt, left_width, right_width, logger):
    ysize = len(y)
    # covered_range cannot overlap the other peak ranges
    left_bound = 0 if i == 0 else peaks[i-1][-1]
    f_ = max(left_bound, pt - left_width*3)
    ex_slice = slice(f_, pt-left_width)
    r_value = linregress(x[ex_slice], y[ex_slice])[2]
    # print([i], "left extended r_value=", r_value)
    if r_value < 0.6:   # 0.569 for 20191006_proteins5 Xray
        # too long extension mus be fixed
        # as in 20191006_proteins5
        n = np.argmax(y[ex_slice])
        start = f_ + n
        new_f_ = start + np.argmin(y[start:pt-left_width])
        logger.info("[%d] covered_range left limit recomputed from %d to %d due to r_value=%.3g", i, f_, new_f_, r_value)
        f_ = new_f_

    right_bound = ysize if i == len(peaks)-1 else peaks[i+1][0]
    t_ = min(right_bound, pt + right_width*3)
    ex_slice = slice(pt+right_width, t_)
    r_value = linregress(x[ex_slice], y[ex_slice])[2]
    # print([i], "right extended r_value=", r_value)
    if r_value > -0.6:
        # too long extension mus be fixed
        n = np.argmax(y[ex_slice])
        start = pt + right_width
        new_t_ = start + np.argmin(y[start:start+n+1])
        logger.info("[%d] covered_range right limit recomputed from %d to %d due to r_value=%.3g", i, t_, new_t_, r_value)
        t_ = new_t_

    return f_, t_

class RobustPeaks:
    def __init__(self, x, y, debug=False):
        self.logger = logging.getLogger(__name__)
        self.x = x
        self.y = y
        self.max_y = np.max(y)
        self.peaks, self.ignore_info = recognize_peaks_roughly(x, y, logger=self.logger, return_ignore_info=True, debug=debug)

    def get_peaks(self):
        return self.peaks

    def get_limits_roughly(self, num_sigmas):
        epeaks = self.get_peaks()
        limit, pt, _ = epeaks[0]
        left_limit = self.get_left_sigma_limit(num_sigmas, limit, pt)

        _, pt, limit = epeaks[-1]
        right_limit = self.get_right_sigma_limit(num_sigmas, limit, pt)
        if self.ignore_info is not None:
            right_limit = min(right_limit, self.ignore_info.slice_.start)
        return left_limit, right_limit

    def get_left_sigma_limit(self, num_sigmas, limit, pt):
        y = self.y
        i = limit
        while i > 0 and y[i-1] < y[i]:
            i -= 1
        width = pt - i
        limit = pt - num_sigmas*(width//2)
        if limit > self.x[0] and self.y[limit]/self.max_y > MIN_HEIGHT_RATIO:
            # as in 20200630_11 Xray
            try:
                sy = smooth(self.y[0:limit])
                limit = np.where(sy < self.max_y * EXTEND_HEIGHT_RATIO)[0][-1]
            except:
                log_exception(self.logger, "extending left_sigma_limit failed: ")
                limit = self.x[0] 
        return max(self.x[0], limit)

    def get_right_sigma_limit(self, num_sigmas, limit, pt):
        y = self.y
        i = limit
        N = len(self.x) - 1
        while i < N and y[i] > y[i+1]:
            i += 1
        width = i - pt
        limit = pt + num_sigmas*(width//2)
        if limit < self.x[-1] and self.y[limit]/self.max_y > MIN_HEIGHT_RATIO:
            # not yet acknowledged
            self.logger.warning("right_sigma_limit might better be extended.")
        return min(self.x[-1], limit)

def demo(name, curve, debug=False):
    from time import time
    from molass_legacy.Elution.CurveUtils import simple_plot

    x = curve.x
    y = curve.y
    t0 = time()
    rp = RobustPeaks(x, y, debug=debug)
    peaks = rp.get_peaks()
    limits = rp.get_limits_roughly(10)
    print("it took %.3g seconds" % (time()-t0))

    plt.push()
    fig, ax = plt.subplots()
    ax.set_title("Demo Plot for " + name)
    simple_plot(ax, curve, legend=False)
    for k, p in enumerate(peaks):
        print([k], p)
        ax.plot(p, y[p], 'o')

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, ymax)
    for p in limits:
        ax.plot([p, p], [ymin, ymax], color='yellow')

    fig.tight_layout()
    plt.show()
    plt.pop()

    return limits



"""
    CustomSigmoid.py


"""
import numpy as np
from scipy.optimize import curve_fit
import molass_legacy.KekLib.DebugPlot as plt
from .DiniSigmoid import dini_sigmoid, dini_sigmoid_inv

HW = 50

def custom_sigmoid(x, L, x0, k, b):
    return L*dini_sigmoid(x - x0, k) + b

def custom_sigmoid_inv(y, L, x0, k, b):
    """
    y = L*dini_sigmoid(x - x0, k) + b
    dini_sigmoid(x - x0, k) = (y - b)/L
    x - x0 = dini_sigmoid_inv((y - b)/L, k)
    x = dini_sigmoid_inv((y - b)/L, k) + x0
    """
    return dini_sigmoid_inv((y - b)/L, k) + x0

def get_safer_point_impl(x, a_y, ratio, a_popt, debug=False):
    L, x0, k, b, s1, s2 = a_popt

    nx = x - x0
    y = a_y - np.hstack([s1*nx[nx < 0], s2*nx[nx >= 0]])

    limits = [int(round(p)) for p in [x0 - HW, x0 + HW]]
    slice_ = slice(max(0, limits[0]), min(len(x), limits[1]))
    x_ = x[slice_]
    y_ = y[slice_]
    xlength = x_[-1] - x_[0]
    wx_ = (x_ - x0)*2/xlength

    # k0 = -0.95 if L > 0 else 0.95
    p0 = (L/2, 0, -0.95, b + L/2)
    popt, pcov = curve_fit(custom_sigmoid, wx_, y_, p0)
    print("popt=", popt)
    assert abs(popt[2]) < 1     # k < -1 occured in Matsumura, which ended up in abnormal result

    if debug:
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(wx_, y_)
        ax.plot(wx_, custom_sigmoid(wx_, *popt))
        plt.show()
        plt.pop()

    L_, x0_, k_, b_ = popt
    y1 = b_ - L_ + L_ * 2 * ratio   # note that L ～ 2L_, the ratio must be scaled from (0, 1) to (-1, 1)
    wx1 = custom_sigmoid_inv(y1, *popt)

    if debug:
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(wx_, y_)
        ax.plot(wx_, custom_sigmoid(wx_, *popt))
        xmin, xmax = ax.get_xlim()
        ax.set_xlim(xmin, xmax)
        ax.plot([xmin, xmax], [y1, y1])
        ax.plot(wx1, y1, "o", color="red")
        plt.show()
        plt.pop()

    x1 = wx1/2*xlength + x0

    if debug:
        from .Sigmoid import sigmoid
        a_popt_ = a_popt[0:4]
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, sigmoid(x, *a_popt_))
        ax.plot(x_, custom_sigmoid(wx_, *popt))
        ax.plot(x1, y1, "o", color="red")
        plt.show()
        plt.pop()

    s = s1 if x1 < x0 else s2
    r_y1 = y1 - s*(x1 - x0)

    return x1, r_y1



# coding: utf-8
"""
    SvdDataSaver.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_savetxt

class SvdDataSaverDialog(Dialog):
    def __init__(self, parent, data_name, svd_results, num_components, location=None):
        self.svd_results = svd_results
        self.default_num = num_components
        self.data_name = data_name
        self.num_components = num_components
        Dialog.__init__(self, parent, "Svd Results Saver for " + data_name, visible=False, location=location)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        label = Tk.Label(iframe, text='Enter your output folder and press "Save" button.')
        label.pack()

        space = Tk.Frame(iframe, height=10)
        space.pack()

        input_frame = Tk.Frame(iframe)
        input_frame.pack()

        self.out_folder = Tk.StringVar()
        data_type = self.data_name.split(' ')[0].lower()
        default_folder = os.path.join(get_setting('analysis_folder'), 'svd-results' + '/' + data_type).replace('\\', '/')
        self.out_folder.set(default_folder)

        entry = FolderEntry(input_frame, textvariable=self.out_folder, width=60)
        entry.pack(side=Tk.LEFT)

        space = Tk.Frame(input_frame, width=20)
        space.pack(side=Tk.LEFT)

        label = Tk.Label(input_frame, text="number of components")
        label.pack(side=Tk.LEFT)

        self.num_components = Tk.IntVar()
        self.num_components.set(self.default_num)
        spinbox = Tk.Spinbox( input_frame, textvariable=self.num_components,
                                from_=1, to=10, increment=1,
                                justify=Tk.CENTER, width=6 )
        spinbox.pack(side=Tk.LEFT)

        space = Tk.Frame(iframe, height=10)
        space.pack()

        self.status = Tk.StringVar()
        self.status_label = Tk.Label(iframe, textvariable=self.status)
        self.status_label.pack()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Save", width=10, command=self.save, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def save(self):
        out_folder = self.out_folder.get()

        if not os.path.exists(out_folder):
            mkdirs_with_retry(out_folder)

        x, U, s, V = self.svd_results
        n = self.num_components.get()

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            fig = plt.figure()
            ax = fig.gca()
            axt = ax.twinx()
            ax.plot(U[:,0:n], label='U')
            ax.plot(V[:,0:n], label='V')
            ax.legend()
            plt.show()

        x_name = "wavelengths.dat" if self.data_name.find("UV") >= 0 else "qvector.dat"
        for data, name in zip([x, U[:,0:n], s[0:n], V[:, 0:n]], [x_name, "U.dat", "s.dat", "V.dat"]):
            np_savetxt(os.path.join(out_folder, name), data)

        self.status.set("your svd results have been successfully saved.")



"""
    GuinierTools.CpdDecompDirect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from GuinierTools.RgCurveUtils import get_connected_curve_info
from molass_legacy.Models.Stochastic.DispersiveMonopore import guess_params_using_moments

MIN_PROP = 0.03

def compute_proportions(x, y, end_points):
    areas = []
    for i in range(len(end_points)-1):
        j = bisect_right(x, end_points[i])
        k = bisect_right(x, end_points[i+1])
        area = np.sum(y[j:k])
        areas.append(area)
    areas = np.array(areas)
    props = areas/np.sum(areas)
    props[props < MIN_PROP] = MIN_PROP
    return props/np.sum(props)

def cpd_direct_impl(editor, debug=False):
    from importlib import reload
    import GuinierTools.CpdDecompUtils
    reload(GuinierTools.CpdDecompUtils)
    from GuinierTools.CpdDecompUtils import compute_end_points
    from Estimators.SdmEstimatorImpl import guess_exec_spec, edit_to_full_sdmparams

    print("cpd_direct_impl")

    x, y, peaks, baseline = editor.xr_draw_info
    rg_curve = editor.dsets[1]

    x_, y_, rgv, qualiteis, valid_bools = get_connected_curve_info(rg_curve)

    nc = editor.get_n_components() - 1
    end_points = compute_end_points(nc, x_, rgv)
    print("end_points", end_points)
    trs = (end_points[0:-1] + end_points[1:])/2
    rg_params = []
    moments_list = []
    width_scale = 0.9     # 
    for i, tr in enumerate(trs):
        j = bisect_right(x_, tr)
        rg_params.append(rgv[j])
        M1 = tr
        M2 = ((tr - end_points[i])*width_scale)**2
        M3 = 0
        moments_list.append((M1, M2, M3))
    
    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("cpd_direct_impl: moments")
            ax.plot(x, y)
            for M in moments_list:
                ax.axvline(x=M[0], color='green')
                s = np.sqrt(M[1])
                ax.axvspan(M[0]-s, M[0]+s, color='green', alpha=0.2)
            fig.tight_layout()
            ret = plt.show()
            if not ret:
                return

    rg_params = np.array(rg_params)
    qualities = np.ones(nc)
    props = compute_proportions(x, y, end_points)
    print("props=", props)

    in_folder = get_setting("in_folder")
    if in_folder.find("20240730") > 0 and nc == 4:
        unreliable_indeces = [3]
    else:
        unreliable_indeces = []

    # exec_spec = guess_exec_spec(peak_rgs, props, qualities)
    exec_spec = {
        "unreliable_indeces": unreliable_indeces,
        "poresize_bounds": (70, 80),
    }

    ret = guess_params_using_moments(x, y, moments_list, rg_params, qualities, props, exec_spec=exec_spec, debug=False)
    if ret is None:
        return
    
    sdm_params, corrected_rgs, bounds = ret
    optimizer = editor.optimizer

    init_params = edit_to_full_sdmparams(editor, sdm_params, corrected_rgs, optimizer.uv_curve, debug=False)
    if init_params is None:
        return
    
    if True:
        with plt.Dp(button_spec=["OK", "Cancel"]):
            fig, ax = plt.subplots()
            ax.set_title("cpd_direct_impl: result")
            axt = ax.twinx()
            axt.grid(False)
            axes = (None, ax, None, axt)
            optimizer.objective_func(init_params, plot=True, axis_info=(fig, axes))
            fig.tight_layout()
            ret = plt.show()
            if not ret:
                return

    editor.draw_scores(init_params, create_new_optimizer=False)


# coding: utf-8
"""
    MatsukenGmm.py.

    adapted from:
        EMアルゴリズム徹底解説
        https://qiita.com/kenmatsu4/items/59ea3e5dfa3d4c161efb

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
import numpy.random as rd
import scipy as sp
from scipy import stats as st
from collections import Counter

import matplotlib
from matplotlib import font_manager
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rc
import matplotlib.animation as ani
plt.rcParams["patch.force_edgecolor"] = True
#rc('text', usetex=True)
import seaborn as sns
sns.set(style="whitegrid", palette="muted", color_codes=True)
sns.set_style("whitegrid", {'grid.linestyle': '--'})
red = sns.xkcd_rgb["light red"]
green = sns.xkcd_rgb["medium green"]
blue = sns.xkcd_rgb["denim blue"]

c = ['r', 'g', 'b']

class GMM:
    def __init__(self, K, max_iter):
        self.K = K
        self.max_iter = max_iter

    def fit(self, X, plot=False):
        self.data = X
        # print('non finite=', X[np.logical_not(np.isfinite(X))])
        self.N, self.D = X.shape
        self.initilize(plot=plot)
        if not plot:
            self.iterate()
        self.weights_ = self.pi
        self.means_ = self.mu.flatten()
        self.covariances_ = self.sigma.flatten()

    def initilize(self, plot=False):
        data = self.data
        K = self.K

        # initialize pi
        pi = np.zeros(K)
        for k in range(K):
            if k == K-1:
                pi[k] = 1 - np.sum(pi)
            else:
                pi[k] = 1/K
        print('init pi:', pi)

        # initialize mu
        mu_list = []
        lim_list = []
        for d in range(self.D):
            max_x, min_x = np.max(data[:,d]), np.min(data[:,d])
            mu_list.append(rd.uniform(low=min_x, high=max_x, size=K))
            lim_list.append([min_x, max_x])

        mu = np.array(mu_list).T
        print('init mu:\n', mu)

        if plot:
            self.xlim = lim_list[0]
            self.ylim = lim_list[1]

            # visualize for check
            #plt.figure(figsize=(12,8))
            plt.figure(figsize=(12, 5))
            plt.scatter(data[:,0], data[:,1], s=30, color='gray', alpha=0.5, marker="+")

            for i in range(3):
                plt.scatter([mu[i, 0]], [mu[i, 1]], color=c[i], marker='o')
            plt.show()

        # initialize sigma
        sigma_list = []
        for k in range(K):
            sigma_list.append(np.eye(self.D)*50)

        sigma = np.asanyarray(sigma_list)

        # calculate likelihood
        likelihood = self.calc_likelihood(data, mu, sigma, pi)
        print('initial sum of log likelihood:', np.sum(np.log(likelihood)))

        print('pi:\n', pi)
        print('mu:\n', mu)
        print('sigma:\n', sigma)
        self.pi = pi
        self.mu = mu
        self.sigma = sigma

        # plt.title('initial state')

    def calc_likelihood(self, data, mu, sigma, pi):
        likelihood = np.zeros((self.N, self.K))
        K = self.K
        for k in range(K):
            likelihood[:, k] = [pi[k]*st.multivariate_normal.pdf(d, mu[k], sigma[k]) for d in data]
        return likelihood

    def calc_prob_gmm(self, data, mu, sigma, pi):
        return [[pi[k]*st.multivariate_normal.pdf(d, mu[k], sigma[k]) for k in range(self.K)] for d in data]

    def print_gmm_contour(self, mu, sigma, pi):
        # display predicted scores by the model as a contour plot
        X, Y = np.meshgrid(np.linspace(*self.xlim), np.linspace(*self.ylim))
        XX = np.array([X.ravel(), Y.ravel()]).T
        Z = np.sum(np.asanyarray(self.calc_prob_gmm(XX, mu, sigma, pi)), axis=1)
        Z = Z.reshape(X.shape)
        CS = plt.contour(X, Y, Z, alpha=0.2, zorder=-100)
        
        plt.title('pdf contour of a GMM')

    def iterate(self):
        pi = self.pi
        mu = self.mu
        sigma = self.sigma
        K = self.K
        N = self.N
        D = self.D
        data = self.data

        print('pi:\n', pi)
        print('mu:\n', mu)
        print('sigma:\n', sigma)

        for step in range(self.max_iter):
            # E step --------------------------------------------------
            likelihood = self.calc_likelihood(data, mu, sigma, pi)
            gamma = (likelihood.T/np.sum(likelihood, axis=1)).T
            N_k = np.array([np.sum(gamma[:,k]) for k in range(K)])

            # M step --------------------------------------------------
            # caluculate pi
            pi = N_k/N

            # calculate mu
            tmp_mu = np.zeros((K, D))

            for k in range(K):
                for i in range(len(data)):
                    tmp_mu[k] += gamma[i, k]*data[i]
                tmp_mu[k] = tmp_mu[k]/N_k[k]
                #print('updated mu[{}]:\n'.format(k) , tmp_mu[k])
            mu_prev = mu.copy()
            mu = tmp_mu.copy()
            #print('updated mu:\n', mu)

            # calculate sigma
            tmp_sigma = np.zeros((K, D, D))

            for k in range(K):
                tmp_sigma[k] = np.zeros((D, D))
                for i in range(N):
                    tmp = np.asanyarray(data[i]-mu[k])[:,np.newaxis]
                    tmp_sigma[k] += gamma[i, k]*np.dot(tmp, tmp.T)
                tmp_sigma[k] = tmp_sigma[k]/N_k[k]

                #print('updated sigma[{}]:\n'.format(k) , tmp_sigma[k])
            sigma = tmp_sigma.copy()

            # calculate likelihood
            prev_likelihood = likelihood
            likelihood = self.calc_likelihood(data, mu, sigma, pi)

            prev_sum_log_likelihood = np.sum(np.log(prev_likelihood))
            sum_log_likelihood = np.sum(np.log(likelihood))
            diff = prev_sum_log_likelihood - sum_log_likelihood
            print([step], diff)

            if np.abs(diff) < 0.0001:
                break

        self.pi = pi
        self.mu = mu
        self.sigma = sigma

    def animate(self, nframe):
        pi = self.pi
        mu = self.mu
        sigma = self.sigma

        K = self.K
        N = self.N
        D = self.D
        data = self.data

        print('nframe:', nframe)
        plt.clf()
        
        if nframe <= 3:
            print('initial state')
            plt.scatter(data[:,0], data[:,1], s=30, color='gray', alpha=0.5, marker="+")
            for i in range(3):
                plt.scatter([mu[i, 0]], [mu[i, 1]], color=c[i], marker='o', edgecolors='k', linewidths=1)
            self.print_gmm_contour(mu, sigma, pi)
            plt.title('initial state')
            return

        # E step ========================================================================
        # calculate responsibility(負担率)
        likelihood = self.calc_likelihood(data, mu, sigma, pi)
        #gamma = np.apply_along_axis(lambda x: [xx/np.sum(x) for xx in x] , 1, likelihood)
        gamma = (likelihood.T/np.sum(likelihood, axis=1)).T
        N_k = [np.sum(gamma[:,k]) for k in range(K)]

        # M step ========================================================================

        # caluculate pi
        pi = np.array(N_k)/N

        # calculate mu
        tmp_mu = np.zeros((K, D))

        for k in range(K):
            for i in range(len(data)):
                tmp_mu[k] += gamma[i, k]*data[i]
            tmp_mu[k] = tmp_mu[k]/N_k[k]
            #print('updated mu[{}]:\n'.format(k) , tmp_mu[k])
        mu_prev = mu.copy()
        mu = tmp_mu.copy()
        #print('updated mu:\n', mu)

        # calculate sigma
        tmp_sigma = np.zeros((K, D, D))

        for k in range(K):
            tmp_sigma[k] = np.zeros((D, D))
            for i in range(N):
                tmp = np.asanyarray(data[i]-mu[k])[:,np.newaxis]
                tmp_sigma[k] += gamma[i, k]*np.dot(tmp, tmp.T)
            tmp_sigma[k] = tmp_sigma[k]/N_k[k]

            #print('updated sigma[{}]:\n'.format(k) , tmp_sigma[k])
        sigma = tmp_sigma.copy()

        # calculate likelihood
        prev_likelihood = likelihood
        likelihood = self.calc_likelihood(data, mu, sigma, pi)
        
        prev_sum_log_likelihood = np.sum(np.log(prev_likelihood))
        sum_log_likelihood = np.sum(np.log(likelihood))
        diff = prev_sum_log_likelihood - sum_log_likelihood

        print('sum of log likelihood:', sum_log_likelihood)
        print('diff:', diff)

        print('pi:', pi)
        print('mu:', mu)
        print('sigma:', sigma)

        self.pi = pi
        self.mu = mu
        self.sigma = sigma

        # visualize
        #plt.figure(figsize=(12,8))
        for i in range(N):
            plt.scatter(data[i,0], data[i,1], s=30, color=gamma[i], alpha=0.5, marker="+")

        for i in range(K):
            ax = plt.axes()
            ax.arrow(mu_prev[i, 0], mu_prev[i, 1], mu[i, 0]-mu_prev[i, 0], mu[i, 1]-mu_prev[i, 1],
                      lw=0.8, head_width=0.02, head_length=0.02, fc='k', ec='k')
            plt.scatter([mu_prev[i, 0]], [mu_prev[i, 1]], color=c[i], marker='o', alpha=0.8)
            plt.scatter([mu[i, 0]], [mu[i, 1]], color=c[i], marker='o', edgecolors='k', linewidths=1)
            #plt.scatter([mu[i, 0]], [mu[i, 1]], color=c[i], marker='o')
        plt.title("step:{}".format(nframe))

        self.print_gmm_contour(mu, sigma, pi)
        # plt.show()
        
        if np.abs(diff) < 0.0001:
            plt.title('likelihood is converged.')
        else:
            plt.title("iter:{}".format(nframe-3))

def generate_demo_data(plot=False):
    n = [200, 150, 150]

    seed = 77

    mu_true = np.asanyarray(
         [[0.2, 0.5],
          [1.2, 0.5],
          [2.0, 0.5]])

    D = mu_true.shape[1]

    sigma_true = np.asanyarray(
            [ [[0.1,  0.085],[ 0.085, 0.1]],
              [[0.1, -0.085],[-0.085, 0.1]],
              [[0.1,  0.085],[ 0.085, 0.1]]
            ])

    rd.seed(seed)
    org_data = None #np.empty((np.sum(n), 3))
    for i in range(3):
        print("check: ", i, mu_true[i], sigma_true[i], np.linalg.det(sigma_true[i]))
        #org_data.append(st.multivariate_normal.rvs(mean=mu[i], cov=sigma[i], size=n[i]))
        if org_data is None:
            org_data = np.c_[st.multivariate_normal.rvs(mean=mu_true[i], cov=sigma_true[i], size=n[i]), np.ones(n[i])*i]
        else:
            org_data = np.r_[org_data, np.c_[st.multivariate_normal.rvs(mean=mu_true[i], 
                                                                        cov=sigma_true[i], 
                                                                        size=n[i]), np.ones(n[i])*i]]
    if plot:
        # plot generated data

        plt.figure(figsize=(12, 5))
        for i in range(3):
            plt.scatter(org_data[org_data[:,2]==i][:,0], org_data[org_data[:,2]==i][:,1], s=30, color=c[i], alpha=0.5)

    # drop true cluster label
    data = org_data[:,0:2].copy()
    return data, mu_true.shape[0]

def blog_demo(plot=False):
    data, K = generate_demo_data(plot)

    if plot:
        plt.show()

    gmm = GMM(K, 1000)
    gmm.fit(data, plot=plot)

    if plot:
        plt.show()

        fig = plt.figure(figsize=(12,5))
        anim = ani.FuncAnimation(fig, gmm.animate, frames=48)
        # anim.save('gmm_anim.gif', writer='imagemagick', fps=3, dpi=128)
        plt.show()



"""
    Trimming.SigmoidApplicability.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct

def check_applicability(x, y, debug=False):
    logger = logging.getLogger(__name__)

    # L ,x0, k, b, s1, s2
    safe_params = 0, 0, 1, 0, 0, 0

    if debug:
        from molass_legacy.Trimming.Sigmoid import ex_sigmoid
        y_ = ex_sigmoid(x, *safe_params)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("check_applicability")
            ax.plot(x, y, label="data")
            ax.plot(x, y_, label="safe sigmoid")
            fig.tight_layout()
            plt.show()

    judge = False
    info = Struct(safe_params=safe_params)
    logger.warning("check_applicability: not implemented yet")
    return judge, info



"""
    Estimators.EdmEstimator.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.PeProgressConstants import MAXNUM_STEPS, STOCH_INIT_STEPS
from .EghEstimator import EghEstimator

class EdmEstimator(EghEstimator):
    def __init__(self, editor, n_components):
        self.n_components = n_components
        EghEstimator.__init__(self, editor)

    def estimate_params(self, debug=False):
        if debug:
            from importlib import reload
            import Models.RateTheory.EDM
            reload(Models.RateTheory.EDM)
        from molass_legacy.Models.RateTheory.EDM import guess_multiple_impl, edm_impl

        init_xr_params, init_xr_baseparams, temp_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, seccol_params = self.estimate_egh_params()

        editor = self.editor
        progress = MAXNUM_STEPS - STOCH_INIT_STEPS
        editor.update_status_bar("Estimating EDM initial parameters.")

        nc = self.n_components - 1   # num components without baseline

        uv_curve, xr_curve = self.ecurves

        x = xr_curve.x
        y = xr_curve.y

        xr_params = guess_multiple_impl(x, y, nc, debug=debug)
        uv_w = np.array([uv_curve.max_y/xr_curve.max_y] * nc)

        progress += 1
        editor.pbar["value"] = progress
        editor.update()

        baseline_type = get_setting("unified_baseline_type")
        if baseline_type >= 2:
            # recompute uv_baseparams
            # to be implemented

            uv_base_params = init_uv_baseparams
        else:
            uv_base_params = init_uv_baseparams

        Tz = np.average(xr_params[:,0])
        init_params = np.concatenate([xr_params.flatten(), init_xr_baseparams, temp_rgs, init_mapping, uv_w, uv_base_params, init_mappable_range, [Tz]])

        progress += 1
        editor.pbar["value"] = progress
        editor.update_status_bar("EDM initial parameters are ready.")

        return init_params

def onthefly_test(editor):
    optimizer = editor.optimizer
    n_components = optimizer.params_type.n_components
    estimator = EdmEstimator(editor, n_components)
    print("estimating...")
    init_params = estimator.estimate_params(debug=True)
    print("done.")
    if init_params is None:
        return
    
    # optimizer.params_type.set_estimator(estimator)
    def draw_params(params, fig, axes):
        optimizer.objective_func(params, plot=True, axis_info=(fig, axes))

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        axt = ax2.twinx()
        axt.grid(False)
        fig.suptitle("EdmEstimator onthefly_test at PeakEditor")
        draw_params(init_params, fig, (ax1, ax2, ax3, axt))
        fig.tight_layout()
        ret = plt.show()
    if not ret:
        print("debug done.")
        return

    editor.draw_scores(init_params, create_new_optimizer=False)
    print("redraw done.")



# coding: utf-8
"""
    DriftLinearModel..py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy as np
from mpl_toolkits.mplot3d   import Axes3D
from matplotlib.collections import PolyCollection
from LmfitThreadSafe        import minimize, Parameters
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from CanvasFrame            import CanvasFrame
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from DataModels             import _GuinierPorod

USE_GUNIER_POROD_MODEL  = True
if USE_GUNIER_POROD_MODEL:
    from DataModels         import _GuinierPorod

TO_USE_QVECTOR_RATIO    = 0.3

LPM_BASE_SHIFT  = 0

class DriftLinearModel:
    def __init__( self, qvector, num_elutions ):
        qstop   = max( 250, int( len(qvector)*TO_USE_QVECTOR_RATIO ) )
        self.qvec   = qvector[0:qstop]
        self.num_elutions = num_elutions
        self.xx     = self.qvec**2
        self.jvec   = np.arange( num_elutions )
        self.fitted_indeces     = None
        self.fitted_baselines   = None

    def simulate( self, params ):
        self.params = params
        Rg      = params[ 'Rg' ]
        Gstart  = params[ 'Gstart' ]
        Gstop   = params[ 'Gstop' ]
        base_pos    = params[ 'base_pos' ]

        A   = Gstart
        B   = ( Gstop - Gstart ) / self.jvec[-1]
        Gb  = A + B*base_pos

        if USE_GUNIER_POROD_MODEL:
            d   = params[ 'd' ]
            Zb  = _GuinierPorod( Gb, Rg, d, self.qvec )
        else:
            z_  = np.exp( -Rg**2 * self.xx/3 )
            Zb  = Gb * z_

        z_list = []
        for j in self.jvec:
            Gj  = A + B*j
            if USE_GUNIER_POROD_MODEL:
                z   = _GuinierPorod( Gj, Rg, d, self.qvec )
            else:
                z   = Gj * z_
            z_list.append( z)

        return Zb, np.array( z_list )

    def guinier_porod_fit( self, q_, y_, sign, particle=None ):

        params = Parameters()
        if particle is None:
            G_init = 1.0
            Gb_init = G_init * ( 10 + sign*2)/10
            params.add( 'Gb',   value=G_init,   min=1e-6,   max=10 )
            params.add( 'G',    value=Gb_init,  min=1e-6,   max=10 )
            params.add( 'Rg',   value=100,  min=30,     max=1000 )
            params.add( 'd',    value=2.5,  min=1,      max=4 )
        else:
            Gb_init  = particle['Gb']
            params.add( 'Gb',   value=Gb_init, vary=False )
            # params.add( 'Gb',   value=Gb_init, min=Gb_init*0.1, max=Gb_init*10)
            G_init  = particle['G']
            params.add( 'G',    value=G_init, min=G_init*0.1, max=G_init*10 )
            params.add( 'Rg',   value=particle['Rg'], vary=False )
            params.add( 'd',    value=particle['d'], vary=False )

        def objective_func( params ):
            Gb  = params['Gb']
            G   = params['G']
            Rg  = params['Rg']
            d   = params['d']
            yb  = _GuinierPorod( Gb, Rg, d, q_ )
            y   = _GuinierPorod( G - LPM_BASE_SHIFT, Rg, d, q_ )
            return ( y - yb ) - y_

        result  = minimize( objective_func, params, args=() )
        ret_params = {}
        for key in [ 'Gb', 'G', 'Rg', 'd']:
            ret_params[key] = result.params[key].value

        if False:
            from CanvasDialog   import CanvasDialog
            Gb  = ret_params['Gb']
            G   = ret_params['G']
            Rg  = ret_params['Rg']
            d   = ret_params['d']
            q1  = 1/Rg * np.sqrt( 3*d/2 )
            # print( 'sign=', sign, 'q1=', q1 )

            def plot_func( fig ):
                ax1 = fig.add_subplot( 121 )
                ax2 = fig.add_subplot( 122 )
                yb  = _GuinierPorod( Gb, Rg, d, q_ )
                ax1.plot( q_, yb, label='yb' )
                y   = _GuinierPorod( G, Rg, d, q_ )
                ax1.plot( q_, y, label='y' )
                ax2.plot( q_, y_, label='y_' )
                ax2.plot( q_, yb - y, label='yb - y' )
                for ax in [ax1, ax2]:
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim( ymin, ymax )
                    ax.plot( [q1, q1], [ymin, ymax], ':', color='black' )
                    ax.legend()
                fig.tight_layout()

            dialog = CanvasDialog( "Debug: q_, y_", adjust_geometry=True )
            dialog.show( plot_func, figsize=(12, 6), toolbar=True )

        return ret_params

    def compute_initial_values( self, q_, bl_array ):
        curve1  = bl_array[:,0]
        curve2  = bl_array[:,-1]
        min0    = min( curve1[0], curve2[0] )
        max0    = max( curve1[0], curve2[0] )
        sign    = 0
        if min0 > 0:
            sign    = -1
        if max0 < 0:
            sign    = +1
        params1 = self.guinier_porod_fit( q_, curve1, sign )
        params2 = self.guinier_porod_fit( q_, curve2, sign )

        print( 'params1=', params1 )
        print( 'params2=', params2 )
        Gb  = np.sqrt( params1['Gb'] * params2['Gb'] )
        Rg  = np.sqrt( params1['Rg'] * params2['Rg'] )
        d   = np.sqrt( params1['d'] * params2['d'] )

        particle = {}
        particle['Gb'] = Gb
        particle['G'] = params1['G']
        particle['Rg'] = Rg
        particle['d'] = d

        params1 = self.guinier_porod_fit( q_, curve1, sign, particle=particle )

        particle['G'] = params2['G']
        params2 = self.guinier_porod_fit( q_, curve2, sign, particle=particle )
        print( 'params1=', params1 )
        print( 'params2=', params2 )
        G1  = params1['G']
        G2  = params2['G']

        return Gb, G1, G2, Rg, d

    def fit( self, indeces, baselines ):
        self.fitted_indeces     = indeces
        self.fitted_baselines   = baselines

        q_  = self.qvec[indeces]

        Gb_init, G1_init, G2_init, Rg_init, d_init = self.compute_initial_values( q_, np.array( baselines ) )

        qq_ = q_**2

        params = Parameters()

        if True:
            Rg_init     = Rg_init
            Gstart_init = G1_init
            Gstop_init  = G2_init
            slope = (G2_init - G1_init)/(self.num_elutions - 1)
            base_pos_init   = ( Gb_init - G1_init ) / slope
            # params.add( 'Rg',   value=Rg_init, vary=False )
            params.add( 'Rg',       value=Rg_init,      min=Rg_init*0.5,    max=Rg_init*5 )
        else:
            Rg_init     = 250
            Gstart_init = 1.0
            Gstop_init  = 0.5
            base_pos_init   = -50
            params.add( 'Rg',       value=Rg_init,      min=Rg_init*0.5,    max=Rg_init*5 )
        params.add( 'Gstart',   value=Gstart_init,  min=0.2,              max=Gstart_init*2 )
        params.add( 'Gstop',    value=Gstop_init,   min=0.2,              max=Gstop_init*2 )
        params.add( 'base_pos',     value=base_pos_init,    min=-len(self.jvec)*0.5,    max=len(self.jvec)*1.5 )

        if USE_GUNIER_POROD_MODEL:

            params.add( 'd',    value=d_init,      min=1.0,    max=4.0 )
            w_   = self.jvec / self.jvec[-1]

            def objective_func( params ):
                Rg      = params['Rg']
                Gstart  = params['Gstart']
                Gstop   = params['Gstop']
                base_pos    = params['base_pos']
                d       = params['d']

                Zs  = _GuinierPorod( Gstart - LPM_BASE_SHIFT, Rg, d, q_ )
                Zp  = _GuinierPorod( Gstop - LPM_BASE_SHIFT,  Rg, d, q_ )

                A   = Gstart
                B   = ( Gstop - Gstart ) / self.jvec[-1]
                Gb  = A + B*base_pos
                Zb  = _GuinierPorod( Gb, Rg, d, q_ )

                diff    = np.zeros( len(self.jvec) )
                for i, baseline in enumerate( baselines ):
                    diff +=  np.abs( ( Zs[i]*(1 - w_) + Zp[i]*w_ - Zb[i] ) - baseline )

                return diff

        else:
            def objective_func( params ):
                Rg      = params['Rg']
                Gstart  = params['Gstart']
                Gstop   = params['Gstop']
                base_pos    = params['base_pos']

                z_  = np.exp( -Rg**2 * qq_/3 )

                A   = Gstart
                B   = ( Gstop - Gstart ) / self.jvec[-1]
                Gb  = A + B*base_pos
                Zb  = Gb * z_

                diff    = np.zeros( len(self.jvec) )
                for i, baseline in enumerate( baselines ):
                    diff +=  np.abs( ( z_[i] * (B * self.jvec + A) - Zb[i] ) - baseline )

                return diff

        result  = minimize( objective_func, params, args=() )

        self.params = {}
        for key in [ 'Rg', 'Gstart', 'Gstop', 'base_pos' ]:
            self.params[key] = result.params[key].value

        self.params['slope'] = ( self.params['Gstop'] - self.params['Gstart'] ) / self.jvec[-1]
        self.params['intercept'] = self.params['Gstart']
        self.params['Gbg'] = self.params['intercept'] + self.params['slope'] * self.params['base_pos' ]

        if USE_GUNIER_POROD_MODEL:
            self.params['d'] = result.params['d'].value

    def draw( self, axes, Zb, data ):
        ax1, ax2 = axes

        ax1.set_title( "Before Background Subtraction" )
        ax2.set_title( "After Background Subtraction" )
        for ax in axes:
            ax.set_xlabel( 'Q' )
            ax.set_ylabel( 'Elution' )
            ax.set_zlabel( 'Intensity' )

        sub_list = []   
        for curve in data:
            sub_list.append( curve - Zb )

        base_pos = self.params[ 'base_pos' ]
        y   = np.ones( len(self.qvec) ) * base_pos

        ax1.plot( self.qvec, y, Zb, color='green', label='Background' )
        ax2.plot( self.qvec, y, np.zeros( len(self.qvec) ), color='green', label='Background' )

        color   = 'red'
        sub_data    = np.array( sub_list )

        def is_to_make_polygon( k ):
            return k % 10 == 0

        poly_alpha = 0.3

        verts1 = []
        verts2 = []
        zs_list = []

        for i, q in enumerate( self.qvec ):
            x   = np.ones(len(self.jvec)) * q
            z1  = data[:,i]
            z2  = sub_data[:,i]

            if is_to_make_polygon( i ):
                zs_list.append( q )
                zb_val  = Zb[i]

                if base_pos < 0 or base_pos >= len(self.jvec):
                    ax1.plot( [q, q], [base_pos, 0], [zb_val, zb_val], ':', color='green', alpha=poly_alpha )
                    ax2.plot( [q, q], [base_pos, 0], [0, 0], ':', color='green', alpha=poly_alpha )

                je  = self.jvec[-1]
                zb  = np.ones( len(self.jvec) ) * zb_val
                ax1.plot( x, self.jvec, zb, color='green', alpha=poly_alpha )
                yp  = [ 0, 0, je, je ]
                zp  = [ zb_val, z1[0], z1[-1], zb_val ]
                verts1.append( list( zip( yp, zp ) ) )

                zb = np.zeros( len(self.jvec) )
                ax2.plot( x, self.jvec, zb, color='green', alpha=poly_alpha )
                zp  = [ 0, z2[0], z2[-1], 0 ]
                verts2.append( list( zip( yp, zp ) ) )

            if i == 0:
                label = 'Stain Intensity'
            else:
                label = None

            ax1.plot( x, self.jvec, z1, color=color, alpha=0.2, label=label )
            ax2.plot( x, self.jvec, z2, color=color, alpha=0.2, label=label )

        if self.fitted_indeces is not None:
            for k, i in enumerate(self.fitted_indeces):
                q   = self.qvec[i]
                x   = np.ones(len(self.jvec)) * q
                z   = self.fitted_baselines[k]
                ax2.plot( x, self.jvec, z, color='yellow', alpha=0.2 )

        poly1 = PolyCollection(verts1)
        poly1.set_alpha(0.2)
        ax1.add_collection3d(poly1, zs=zs_list, zdir='x')

        poly2 = PolyCollection(verts2)
        poly2.set_alpha(0.2)
        ax2.add_collection3d(poly2, zs=zs_list, zdir='x')

        if USE_GUNIER_POROD_MODEL:
            # TODO: plot q1-plane
            pass

        if False:
            zmin1, zmax1 = ax1.get_zlim()
            zmin2, zmax2 = ax2.get_zlim()

            zmin_ = min( zmin1, zmin2 )
            zmax_ = max( zmax1, zmax2 )

            ax1.set_zlim( zmin_, zmax_ )
            ax2.set_zlim( zmin_, zmax_ )

        ax1.legend()
        ax2.legend()

class DriftLinearModelDialog( Dialog ):
    def __init__( self, parent, model, init_params ):
        self.parent = parent
        self.model  = model
        self.init_params = init_params

    def show( self ):
        title = "Linear Drift Simulator"
        Dialog.__init__(self, self.parent, title )

    def body( self, body_frame ):
        # tk_set_icon_portable( self )

        figsize = ( 12, 10 ) if is_low_resolution() else ( 14, 12 )
        self.canvas_frame = canvas_frame = CanvasFrame( body_frame, figsize=figsize )
        canvas_frame.pack()
        fig = self.canvas_frame.fig
        ax1 = fig.add_subplot( 221, projection='3d' )
        ax2 = fig.add_subplot( 222, projection='3d' )
        ax3 = fig.add_subplot( 223, projection='3d' )
        ax4 = fig.add_subplot( 224, projection='3d' )

        fig.tight_layout()

        Zb1, data1  = self.model.simulate( self.init_params )
        self.model.draw( [ ax1, ax2 ], Zb1, data1 )

        copy_params = copy.deepcopy( self.init_params )
        copy_params['Gstart'] += 2.0
        copy_params['Gstop'] += 2.0

        Zb2, data2  = self.model.simulate( copy_params )
        self.model.draw( [ ax3, ax4 ], Zb2, data2 )

        ax1.set_zlim( ax3.get_zlim() )

        self.canvas_frame.show()



"""
    ProgressInfoUtil.py

    Copyright (c) 2017-2024, Masatsuyo Takahashi, KEK-PF
"""

"""
    stream      phase       max
    Base Corr.  0           100

    Guinier     0   autorg  num_files
                1   excel   openpyxl(), com()

    Zero-Ex     0   asc(0)  
                1   desc(0) 
                2   asc(1)  
                3   desc(1) 
                .
                1000  merge   com()
"""
from molass_legacy.SerialAnalyzer.AnalyzerUtil import get_analysis_ranges
from molass_legacy.KekLib.ProgressInfo import put_info
from ZeroExtrapolator import NUM_EXTRAPOLATION_POINTS
from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges

NUM_SHEET_TYPES = 3     # ( ASC,  DATGNOM-ASC ), ( DESC , DATGNOM-DESC ), OVERLAY,
STREAM_BASECOR  = 0
STREAM_GUINIER  = 1
STREAM_ZERO_EX  = 2
STREAM_BASECOR_MAX  = 10

def estimate_gunier( num_files, max_dict={} ):
    max_dict[ (STREAM_GUINIER, 0) ]  = num_files
    max_dict[ (STREAM_GUINIER, 1) ]  = 5
    return max_dict

def extimate_zero_extrapolation( ranges, max_dict={} ):
    num_zx_points = NUM_EXTRAPOLATION_POINTS

    for i, range_ in enumerate( ranges ):
        fromto_list = range_.get_fromto_list()
        lower, middle1 = fromto_list[0]
        if len(fromto_list) > 1:
            middle2, upper = fromto_list[1]
        else:
            middle2 = None
        max_dict[ (STREAM_ZERO_EX, i*NUM_SHEET_TYPES   ) ] = 1 + ( middle1 - lower ) + 1 + num_zx_points + 1 + 1
        if middle2 is not None:
            max_dict[ (STREAM_ZERO_EX, i*NUM_SHEET_TYPES+1 ) ] = 1 + ( upper - middle2 ) + num_zx_points + 1 + 1
        max_dict[ (STREAM_ZERO_EX, i*NUM_SHEET_TYPES+2 ) ] = 1

        # TODO: is this asymmetry ok?

    num_sheets      = len( ranges )*NUM_SHEET_TYPES
    max_dict[ (STREAM_ZERO_EX, 1000) ] = num_sheets
    max_dict[ (STREAM_ZERO_EX, 2000) ] = 1
    # print( 'max_dict=', max_dict )
    return max_dict

def estimate_init_max_dist( num_files, scattering_correction, ranges, zx_flag ):
    print( 'estimate_init_max_dist: scattering_correction=', scattering_correction )

    num_zx_points = NUM_EXTRAPOLATION_POINTS

    stream_basecorr_max = 0 if scattering_correction == 0 else STREAM_BASECOR_MAX
    max_dict = { (STREAM_BASECOR, 0):stream_basecorr_max }

    guinier_max_dict = estimate_gunier( num_files, max_dict=max_dict )

    if ranges is None:
        ranges = get_analysis_ranges()

    if zx_flag:
        applied_ranges, _ = convert_to_paired_ranges( ranges )
        init_max_dict = extimate_zero_extrapolation( applied_ranges, max_dict=guinier_max_dict )
    else:
        init_max_dict = guinier_max_dict

    # print( 'init_max_dict=', init_max_dict )
    return init_max_dict

class ProgressCallback:
    def __init__(self, max_progress, debug=True):
        self.debug = debug
        if debug:
            import logging
            self.logger = logging.getLogger('ProgressCallback')
        self.max_progress   = max_progress

    def __call__( self, i ):
        if i % 100 == 0 or i == self.max_progress:
            if i == self.max_progress:
                progress    = STREAM_BASECOR_MAX
            else:
                # this calculation is not accurate
                progress    = int( STREAM_BASECOR_MAX * i / self.max_progress )
            if self.debug:
                self.logger.info("put_info( (%d, 0), %d )", STREAM_BASECOR, progress)
            put_info( (STREAM_BASECOR, 0), progress )



"""
    SSDC.SsdcAnalysisSdm.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
from importlib import reload
import Models.Stochastic.DispersiveMonopore
reload(Models.Stochastic.DispersiveMonopore)
from molass_legacy.Models.Stochastic.DispersiveMonopore import guess_params_using_moments
import SSDC.SsdcAnalysisUtils
reload(SSDC.SsdcAnalysisUtils)
from SSDC.SsdcAnalysisUtils import plot_ssdc_result

def spike(info_list, num_components, exec_spec):
    print("spike")

    if False:
        exec_spec = {
        # "input_folders" :       [ PyTools + r"\Data\20230706\ALD_OA001",   PyTools + r"\Data\20230706\ALD_OA002"],
        "num_components" :      3,
        "unreliable_indeces" :  [1],
        "poresize_bounds" :      (75, 80),
        "init_N0" :             [50000, 7000],
        }

    names = [folder.split('\\')[-1] for folder, info in info_list]
    print("names=", names)

    rg_info_list = []
    sdm_params_info_list = []
    moments_list = []
    mapping_list = []
    for i, (in_folder, lrfsrc) in enumerate(info_list):
        set_setting('in_folder', in_folder)
        print([i], "getting rg info")
        info = lrfsrc.compute_rgs(want_num_components=num_components, debug=False)
        rg_info_list.append(info)

        print([i], "getting sdm info")
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = info
        peaks = lrfsrc.xr_peaks[indeces]
        egh_moments_list = compute_egh_moments(peaks)
        moments_list.append(egh_moments_list)
        mapping_list.append([M[0] for M in egh_moments_list])
        x = lrfsrc.xr_x
        y = lrfsrc.xr_y
        exec_spec_ = exec_spec.copy()
        exec_spec_["init_N0"] = exec_spec_["init_N0"][i]
        ret = guess_params_using_moments(x, y, egh_moments_list, peak_rgs, qualities, props, exec_spec=exec_spec_, debug=False)
        sdm_params_info_list.append(ret)

    mx = mapping_list[0]
    my = mapping_list[1]
    slope, intercept, r_value, p_value, std_err = linregress(mx, my)
    mapping = slope, intercept

    unreliable_indeces = exec_spec['unreliable_indeces']
    def paired_analysis():
        import Models.Stochastic.PairedAnalysisImpl
        reload(Models.Stochastic.PairedAnalysisImpl)
        from molass_legacy.Models.Stochastic.PairedAnalysisImpl import paired_analysis_impl
        paired_analysis_impl(names, num_components, info_list, rg_info_list,
                             unreliable_indeces, sdm_params_info_list, mapping_list, mapping, moments_list, exec_spec)

    extra_button_specs = [
                ("Paired Analysis", paired_analysis),
                ]

    plot_ssdc_result(names, info_list, rg_info_list, unreliable_indeces,
                     sdm_params_info_list, mapping_list, mapping,
                     extra_button_specs=extra_button_specs)


"""
    Optimizer.Strategies.SdmStrategy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from .BasicStrategy import BasicStrategy

class SdmStrategy(BasicStrategy):
    def __init__(self, **kwargs):
        self.cycle = 7
        self.nc = nc = kwargs.pop("nc", 3)   # excluding baseline

        # cycles 1-2
        xr_start = 0
        uv_start = nc*2 + 2 + 2
        mapping_start = nc*2 + 2
        mapping_indeces = [mapping_start, mapping_start+1]
        self.indeces_list = [list(range(start, start+nc)) + mapping_indeces for start in [xr_start, uv_start]]

        # cycle 3
        rg_start = nc + 2
        self.indeces_list.append(list(range(rg_start, rg_start+nc)))

        # cycle 4
        uv_baseline_start = nc*3 + 2 + 2
        self.indeces_list.append(list(range(uv_baseline_start, uv_baseline_start+7)))

        # cycles 5-6
        # 0, 1, 2,  3,        4,  5
        # N, K, x0, poresize, N0, tI = sdmcol_params
        cp_start = nc*3 + 2 + 7 + 2 + 2
        for list_ in [2, 3, 5], [0, 1, 4]:
            self.indeces_list.append([cp_start + i for i in list_])

    def is_strategic(self, n):
        if self.cycle is None:
            return False
        if n % self.cycle == 0:
            return False
        return True

    def get_indeces_list(self, n):
        r = n % self.cycle
        assert r > 0
        return [self.indeces_list[r-1]]


# coding: utf-8
"""
    EmgEstimation.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
import molass_legacy.KekLib.DebugPlot as plt

def spike():
    N = 1000000

    mu = 5
    sigma = 1.5
    tau = 2

    X = np.random.normal(mu, sigma, N)
    Y = np.random.exponential(tau, N)
    Z = X + Y

    mean = np.mean(Z)
    print('mean=', mean)

    m2 = np.sum((Z - mean)**2)/N
    print('m2=', m2)

    m3 = np.sum((Z - mean)**3)/N
    print('m3=', m3)
    print('tau=', np.power(m3/2, 1/3))
    print('sigma=', np.sqrt(m2 - tau**2))



"""
    Batch/DataWalk.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.SerialAnalyzer.DataUtils import get_pytools_folder, serial_folder_walk
from molass_legacy.Batch.LiteBatch import LiteBatch
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, set_setting

class CallbackWrapper:
    """
    This wrapper adjust the arguments of callback function
    so as to match the signature of the old-style callback function.
    """
    def __init__(self, callback, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.batch = LiteBatch()
        self.callback = callback
        self.ignore_error = kwargs.get("ignore_error", True)
        self.start_folder = kwargs.get("start_folder", None)
        self.counters = kwargs.get("counters", None)
        self.started = False

    def __call__(self, xr_folder, uv_folder=None, plot=False):
        if self.counters is None:
            count = None
        else:
            count = self.counters[0]
            self.counters[0] += 1
        if self.started:
            pass
        else:
            if self.start_folder is not None:
                if xr_folder.find(self.start_folder) < 0:
                    print("skipping", xr_folder)
                    return True, None
            self.started = True

        clear_temporary_settings()
        set_setting("in_folder", xr_folder)
        set_setting("test_pattern", 0)  # this is to avoid the popping up of the dialog box in irregular cases
        try:
            ret = self.callback(xr_folder, self.batch, count)
        except:
            log_exception(self.logger, "Error occurred during processing of %s: " % xr_folder, n=10)
            ret = self.ignore_error
        return ret, None

def do_a_folder(callback, batch, in_folder, count):
    set_setting("in_folder", in_folder)
    callback(in_folder, batch, count)
    
def test_data_walk(callback, **kwargs):
    exact_folder = kwargs.get("exact_folder", None)
    in_folder_file = kwargs.get("in_folder_file", None)
    if exact_folder is None and in_folder_file is None:
        root_folder = kwargs.get("root_folder", None)
        if root_folder is None:
            root_folder = get_pytools_folder()

        wrapper = CallbackWrapper(callback, **kwargs)
        serial_folder_walk(root_folder, wrapper)
    else:
        batch = LiteBatch()       
        if in_folder_file is None:
            # i.e., exact_folder is not None
            import os
            import sys
            this_dir = os.path.dirname( os.path.abspath( __file__ ) )
            sys.path.append( this_dir + '/../../..' )
            from TestEnv import get_data_folder    
            in_folder = get_data_folder(exact_folder)
            do_a_folder(callback, batch, in_folder, 0)
        else:
            with open(in_folder_file) as fh:
                for k, line in enumerate(fh):
                    in_folder = line[:-1].split(',')[0]
                    clear_temporary_settings()
                    set_setting("test_pattern", 0)
                    do_a_folder(callback, batch, in_folder, k)


"""
    OnTheFly.DebugDialog.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
if __name__ == '__main__':
    import os
    import sys
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

from molass_legacy.KekLib.OurTkinter import Tk, Dialog

class DebugDialog(Dialog):
    def __init__(self, parent=None, debug_info=None):
        if parent is None:
            from molass_legacy.KekLib.TkUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        self.debug_info = debug_info
        Dialog.__init__(self, parent, "Debug Dialog", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        label = Tk.Label(body_frame,
                            text="Tell the programmer if you are seeing this dialog.",
                            bg="white", fg="orange", font=('', 16))
        label.pack(padx=20, pady=10)
 
        run_button = Tk.Button(body_frame, text="Run", command=self.run)
        run_button.pack()

    def run(self):
        from importlib import reload
        import OnTheFly.DebugImpl
        reload(OnTheFly.DebugImpl)
        from OnTheFly.DebugImpl import debug_impl
        debug_impl(self)

if __name__ == '__main__':
    dialog = DebugDialog()
    dialog.show()
